import os
import warnings

from marshmallow.warnings import ChangedInMarshmallow3Warning
import marshmallow as mm

from argschema import ArgSchema
from argschema.fields import (
    Boolean, InputDir, InputFile, Float,
    OutputDir, List, Str, Dict)

warnings.simplefilter(
        action='ignore',
        category=ChangedInMarshmallow3Warning)


class MontageSolverSchema(ArgSchema):
    data_dir = InputDir(
        required=False,
        description="directory containing metafile, images, and matches")
    metafile = InputFile(
        required=False,
        description=("fullpath to metafile. Helps in the case of multiple"
                     " metafiles in one directory. data_dir will take "
                     " os.path.dirname(metafile)"))
    output_dir = OutputDir(
        required=False,
        missing=None,
        default=None,
        description="directory for output files")
    read_transform_from = Str(
        required=False,
        missing='metafile',
        default='metafile',
        validator=mm.validate.OneOf(['metafile', 'reffile', 'dict']),
        description="3 possible ways to read in the reference transform")
    ref_transform = InputFile(
        required=False,
        missing=None,
        default=None,
        description="transform json")
    ref_transform_dict = Dict(
        require=False,
        missing=None,
        description="transform in from memory")
    ransacReprojThreshold = Float(
        required=False,
        missing=10.0,
        default=10.0,
        description=("passed into cv2.estimateAffinePartial2D()"
                     "for RANSAC filtering of montage template matches"))
    compress_output = Boolean(
        required=False,
        missing=True,
        default=True,
        description=("tilespecs will be .json or .json.gz"))
    solver_templates = List(
        Str,
        required=True,
        description="input json basenames for the solver args")
    solver_template_dir = InputDir(
        required=True,
        description="location of the templates for the solver")

    @mm.post_load
    def check_solver_inputs(self, data):
        for args in data['solver_templates']:
            argpath = os.path.join(data['solver_template_dir'], args)
            if not os.path.isfile(argpath):
                raise mm.ValidationError(
                        "solver arg file doesn't exist: %s" % argpath)

    @mm.post_load
    def check_metafile(self, data):
        if ('data_dir' not in data) & ('metafile' not in data):
            raise mm.ValidationError(" must specify either data_dir"
                                     " or metafile")



import datetime
import glob
import json
import os
import re
import warnings

from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
import numpy as np

from argschema import ArgSchemaParser
from bigfeta import jsongz
import renderapi

from em_stitch.utils.utils import src_from_xy
from em_stitch.plots.schemas import LensQuiverSchema

warnings.simplefilter(action='ignore', category=FutureWarning)

example = {
        "transform_list": [
            "/data/em-131fs3/lctest/T3_OL5_5pct/20190429113358_reference/0/resolvedtiles.json.gz",
            "/data/em-131fs3/lctest/T3_OL5_5pct/20190429121307_reference/0/resolvedtiles.json.gz",
            "/data/em-131fs3/lctest/T3_OL5_5pct/20190429122415_reference/0/resolvedtiles.json.gz",
            "/data/em-131fs3/lctest/T3_OL5_5pct/20190429134601_reference/0/resolvedtiles.json.gz",
            "/data/em-131fs3/lctest/T3_OL5_5pct/20190429143510_reference/0/resolvedtiles.json.gz",
            "/data/em-131fs3/lctest/T3_OL5_5pct/20190429153110_reference/0/resolvedtiles.json.gz",
            ]
        }


def plot_lens_changes(
        lcs, arrow_scale=10.0, num=1, pdfname='lens_changes.pdf'):
    scale = 1.0 / arrow_scale
    with PdfPages(pdfname) as pdf:
        for lc in lcs:
            tstamp = re.findall('\d+_reference', lc)[0].split('_')[0]
            dt = datetime.datetime.strptime(tstamp, '%Y%m%d%H%M%S')
            dtsf = dt.strftime('%Y-%m-%d %H:%M:%S')
            ddir = os.path.dirname(lc)
            mfile = glob.glob(os.path.join(ddir, '_metadata*.json'))[0]
            with open(mfile, 'r') as f:
                meta = json.load(f)
            obj_focus = meta[0]['metadata']['objective_focus']
            print(obj_focus)
            with open(lc, 'r') as f:
                tf = renderapi.transform.ThinPlateSplineTransform(
                        json=json.load(f))
                sz = tf.srcPts.max(axis=1)
                src = src_from_xy(
                        np.linspace(0, sz[0], 20),
                        np.linspace(0, sz[1], 20))
                dst = tf.tform(src)
                delta = dst - src
                rmax = np.linalg.norm(delta, axis=1).max()
                fig, axes = plt.subplots(
                        1, 1, num=1, clear=True, figsize=(11, 8))
                axes.quiver(
                        src[:, 0], src[:, 1], delta[:, 0], delta[:, 1],
                        angles='xy', scale=scale, scale_units='xy')
                axes.invert_yaxis()
                axes.set_aspect('equal')
                axes.set_title(
                        dtsf + '\n' + lc + "\nfull transform max: %0.1f "
                        "pixels\narrow scale = %0.1f\nobjective focus: %d" %
                        (rmax, arrow_scale, obj_focus))
                pdf.savefig(fig)


def load_transform(path):
    j = jsongz.load(path)

    try:
        return renderapi.transform.ThinPlateSplineTransform(
                json=j)
    except KeyError:
        res = renderapi.resolvedtiles.ResolvedTiles(json=j)
        return res.transforms[0]


def grid_from_tform(tform, xpts=20, ypts=20):
    xmin = tform.srcPts[0, :].min()
    xmax = tform.srcPts[0, :].max()
    ymin = tform.srcPts[1, :].min()
    ymax = tform.srcPts[1, :].max()

    x = np.linspace(xmin, xmax, xpts)
    y = np.linspace(ymin, ymax, ypts)

    return src_from_xy(x, y)


def lens_quiver(ax, tform, src, arrow_scale=1.0, title=None):
    dst = tform.tform(src)
    delta = dst - src

    rmax = np.linalg.norm(delta, axis=1).max()

    scale1 = 1.0 / arrow_scale

    ax.quiver(
            src[:, 0],
            src[:, 1],
            delta[:, 0],
            delta[:, 1],
            angles='xy',
            scale=scale1,
            scale_units='xy')
    if not ax.axes.yaxis_inverted():
        ax.invert_yaxis()
    ax.set_aspect('equal')
    if not title:
        title = ''
    title += "\nfull transform max: %0.1f " % rmax
    title += "pixels\narrow scale = %0.1f" % (1/scale1)
    ax.set_title(title, fontsize=8)


class LensQuiverPlots(ArgSchemaParser):
    default_schema = LensQuiverSchema

    def run(self):
        tforms = [load_transform(p) for p in self.args['transform_list']]

        grid = grid_from_tform(
                tforms[0],
                xpts=self.args['n_grid_pts'],
                ypts=self.args['n_grid_pts'])

        nrow, ncol = self.args['subplot_shape']
        figsize = (10, 10)
        f, a = plt.subplots(
                nrow, ncol, num=self.args['fignum'], figsize=figsize,
                clear=True, sharex=True, sharey=True)
        if not isinstance(a, np.ndarray):
            a = np.array([a])

        iplot = 0
        with PdfPages(self.args['pdf_out']) as pdf:
            for tform, p in zip(tforms, self.args['transform_list']):
                lens_quiver(
                        a[iplot], tform, grid,
                        self.args['arrow_scale'], title=p)
                iplot += 1
                if iplot == (nrow * ncol):
                    pdf.savefig(f)
                    if tform != tforms[-1]:
                        fn = f.number + 1
                        f, a = plt.subplots(
                                nrow, ncol, num=fn, figsize=figsize,
                                clear=True, sharex=True, sharey=True)
                        if not isinstance(a, np.ndarray):
                            a = np.array([a])
                        iplot = 0

        print('wrote %s' % os.path.abspath(self.args['pdf_out']))

        if self.args['show']:
            plt.show()


if __name__ == "__main__":
    qmod = LensQuiverPlots(input_data=example)
    qmod.run()



import copy
import datetime
import logging
import os

from six.moves import urllib

import cv2
import numpy as np
import scipy.optimize
import scipy.sparse as sparse
from scipy.sparse import csr_matrix
from scipy.sparse.linalg import factorized
from scipy.spatial import Delaunay
import triangle

from argschema import ArgSchemaParser
from bigfeta import jsongz
import renderapi

from .schemas import MeshLensCorrectionSchema
from .utils import remove_weighted_matches

try:
    # pandas unique is faster than numpy, use where appropriate
    import pandas
    _uniq = pandas.unique
except ImportError:
    import numpy
    _uniq = numpy.unique


logger = logging.getLogger()
default_logger = logger


class MeshLensCorrectionException(Exception):
    """Exception raised when there is a
    problem creating a mesh lens correction
    """


def condense_coords(matches):
    """restructure point match dictionary structure to Nx2 array

    Parameters
    ----------
    matches : list of dict
        list of match dictionaries in Render format

    Returns
    -------
    coords : numpy.ndarray
        Nx2 array representing matches
    """
    x = []
    y = []
    for m in matches:
        x += m['matches']['p'][0]
        x += m['matches']['q'][0]
        y += m['matches']['p'][1]
        y += m['matches']['q'][1]
    coords = np.transpose(np.vstack((np.array(x), np.array(y))))
    return coords


def smooth_density_legacy(coords, tile_width, tile_height, n):
    """legacy function to homogenize distribution of points within a
        rectangular area by reducing the number of points within
        n**2 equally-sized bounding boxes to
        the minimum number of points in one of those boxes.


    Parameters
    ----------
    coords : numpy.ndarray
        Nx2 numpy array of coordinates to consider
    tile_width : int
        width of rectangular area containing coords
    tile_height : int
        height of rectangular area containing coords
    n : int
        number of subdivisions into which tile_width and tile_height
        should be divided

    Returns
    -------
    smoothed_coords : numpy.ndarray
        Nx2 numpy array of smoothed subset of input coords
    """
    # n: area divided into nxn
    min_count = np.Inf
    for i in range(n):
        r = np.arange(
                (i*tile_width/n),
                ((i+1)*tile_width/n))
        for j in range(n):
            c = np.arange(
                    (j*tile_height/n),
                    ((j+1)*tile_height/n))
            ind = np.argwhere(
                    (coords[:, 0] >= r.min()) &
                    (coords[:, 0] <= r.max()) &
                    (coords[:, 1] >= c.min()) &
                    (coords[:, 1] <= c.max())).flatten()
            if ind.size < min_count:
                min_count = ind.size
    new_coords = []
    for i in range(n):
        r = np.arange(
                (i*tile_width/n),
                ((i+1)*tile_width/n))
        for j in range(n):
            c = np.arange(
                    (j*tile_height/n),
                    ((j+1)*tile_height/n))
            ind = np.argwhere(
                    (coords[:, 0] >= r.min()) &
                    (coords[:, 0] <= r.max()) &
                    (coords[:, 1] >= c.min()) &
                    (coords[:, 1] <= c.max())).flatten()
            a = np.arange(ind.size)
            np.random.shuffle(a)
            ind = ind[a[0:min_count]]
            new_coords.append(coords[ind])
    return np.concatenate(new_coords)


def get_bboxes(tile_width, tile_height, n):
    """get list of bounds for n**2 equally-sized bounding boxes within a
        rectangular bounding box


    Parameters
    ----------
    tile_width : int
        width of rectangular area to divide
    tile_height : int
        height of rectangular area to divide
    n : int
        number of subdivisions into which tile_width and tile_height
        should be divided

    Returns
    -------
    vtxs : list of tuple of numpy.ndarray
        list of min/max tuples of vertices representing bounding boxes
    """
    numX = n
    numY = n
    diffX = (tile_width-1) / numX
    diffY = (tile_height-1) / numY

    squaremesh = np.mgrid[
        0:tile_width-1:numX*1j, 0:tile_height-1:numY*1j].reshape(2, -1).T
    maxpt = squaremesh.max(axis=0)

    vtxs = []
    for pt in squaremesh:
        if np.any(pt == maxpt):
            continue
        vtxs.append((pt, pt + np.array([diffX, diffY])))
    return vtxs


def smooth_density_bbox(coords, tile_width, tile_height, n):
    """homogenize distribution of points within a rectangular area by reducing
        the number of points within n**2 equally-sized bounding boxes to
        the minimum number of points in one of those boxes.


    Parameters
    ----------
    coords : numpy.ndarray
        Nx2 numpy array of coordinates to consider
    tile_width : int
        width of rectangular area containing coords
    tile_height : int
        height of rectangular area containing coords
    n : int
        number of subdivisions into which tile_width and tile_height
        should be divided

    Returns
    -------
    smoothed_coords : numpy.ndarray
        Nx2 numpy array of smoothed subset of input coords
    """
    vtxs = get_bboxes(tile_width, tile_height, n)

    index_arr = np.zeros(coords.shape[0], dtype="int64")
    for ei, (ll, ur) in enumerate(vtxs):
        vi = ei + 1  # 0 is prohibited
        index_arr[np.all((ll <= coords) & (ur >= coords), axis=1)] = vi

    bc = np.bincount(index_arr)
    mincount = bc[1:].min()  # 0 is prohibited

    idxs = _uniq(index_arr)
    new_coords = np.concatenate([
        coords[np.random.choice(np.argwhere(index_arr == idx).flatten(),
                                mincount)] for idx in idxs])

    return new_coords


def smooth_density(coords, tile_width, tile_height, n,
                   legacy_smooth_density=False, **kwargs):
    """homogenize distribution of points within a rectangular area by reducing
        the number of points within n**2 equally-sized bounding boxes to
        the minimum number of points in one of those boxes.


    Parameters
    ----------
    coords : numpy.ndarray
        Nx2 numpy array of coordinates to consider
    tile_width : int
        width of rectangular area containing coords
    tile_height : int
        height of rectangular area containing coords
    n : int
        number of subdivisions into which tile_width and tile_height
        should be divided
    legacy_smooth_density : boolean
        whether to use (slower) legacy code.  Not recommended.

    Returns
    -------
    smoothed_coords : numpy.ndarray
        Nx2 numpy array of smoothed subset of input coords
    """
    if legacy_smooth_density:
        return smooth_density_legacy(coords, tile_width, tile_height, n)
    else:
        return smooth_density_bbox(coords, tile_width, tile_height, n)


def approx_snap_contour(contour, width, height, epsilon=20, snap_dist=5):
    """Approximate a contour within a number of pixels, so it isn't too
    fine in the corner and snap to edges

    Parameters
    ----------
    contour : numpy.ndarray
        Nx2 array input to cv2.approxPolyDP
    width : int
        width to which approximated values are snapped
    height : int
        height to which approximated values are snapped
    epsilon : int
        maximum pixel distance between the original curve and its approximation

    Returns
    -------
    approx : numpy.ndarray
        Polygon approximating the contour
    """
    # approximate contour within epsilon pixels,
    # so it isn't too fine in the corner
    # and snap to edges
    approx = cv2.approxPolyDP(contour, epsilon, True)
    for i in range(approx.shape[0]):
        for j in [0, width]:
            if np.abs(approx[i, 0, 0] - j) <= snap_dist:
                approx[i, 0, 0] = j
        for j in [0, height]:
            if np.abs(approx[i, 0, 1] - j) <= snap_dist:
                approx[i, 0, 1] = j
    return approx


def create_PSLG(tile_width, tile_height, maskUrl):
    """create a PSLG (Planar Straight Line Graph) based on a masked image

    Parameters
    ----------
    tile_width : int
        width of tile for which PSLG will be created
    tile_height : int
        height of tile for which PSLG will be created
    maskUrl : str or None
        file uri to binary mask image

    Returns
    -------
    bbox : dict
        dictionary with keys vertices and segments representing the PSLG
    """
    # define a PSLG for triangle
    # http://dzhelil.info/triangle/definitions.html
    # https://www.cs.cmu.edu/~quake/triangle.defs.html#pslg
    if maskUrl is None:
        vertices = np.array([
                [0, 0],
                [0, tile_height],
                [tile_width, tile_height],
                [tile_width, 0]])
        segments = np.array([
                [0, 1],
                [1, 2],
                [2, 3],
                [3, 0]])
    else:
        mpath = urllib.parse.unquote(
                    urllib.parse.urlparse(maskUrl).path)
        im = cv2.imread(mpath, 0)
        _, contours, _ = cv2.findContours(
                im, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE)
        approx = approx_snap_contour(contours[0], tile_width, tile_height)
        vertices = np.array(approx).squeeze()
        segments = np.zeros((vertices.shape[0], 2))
        segments[:, 0] = np.arange(vertices.shape[0])
        segments[:, 1] = segments[:, 0] + 1
        segments[-1, 1] = 0
    bbox = {}
    bbox['vertices'] = vertices
    bbox['segments'] = segments

    return bbox


def calculate_mesh(a, bbox, target, get_t=False):
    # FIXME target/get_t is ugly
    """triangulate a mesh based on a Planar Straight Line Graph with
    mesh constraints.  Return either the Delaunay triangulation or a
    difference in number of triangles from a target.

    Parameters
    ----------
    a : float
        global maximum triangular area constraint
    bbox : dict
        dictionary with keys vertices and segments representing PSLG
    target : int
        number of triangles to which the triangulation will be compared if
        get_t is False
    get_t : boolean
        whether to return a Delaunay Triangulation rather than a
        comparison to target

    Returns
    -------
    val : scipy.spatial.Delaunay or int
        Delaunay triangulation if get_t is True, otherwise the number of
        vertices in the triangulation subtracted from the target
    """
    t = triangle.triangulate(bbox, 'pqa%0.1f' % a)
    if get_t:
        # scipy.Delaunay has nice find_simplex method,
        # but, no obvious way to iteratively refine meshes, like triangle
        # numbering is different, so, switch here
        return Delaunay(t['vertices'])
    return target - len(t['vertices'])


def force_vertices_with_npoints(area_par, bbox, coords, npts, **kwargs):
    """create a triangular mesh which iteratively attempts to conform to a
        minimum number of points per vertex by adjusting the maximum
        triangle area

    Parameters
    ----------
    area_par : float
        initial maximum triangle area constraint for triangle.triangulate
    bbox : dict
        PSLG bounding box dictionary from :func:create_PSLG
    coords : numpy.ndarray
        Nx2 points
    npts : int
        minimum number of points near each vertex

    Returns
    -------
    t : scipy.spatial.qhull.Delaunay
        triangle mesh with minimum point count near vertices
    area_par : float
        area parameter used to calculate result t
    """
    fac = 1.02
    count = 0
    max_iter = 20
    while True:
        t = calculate_mesh(
                area_par,
                bbox,
                None,
                get_t=True)
        pt_count = count_points_near_vertices(t, coords, **kwargs)
        if pt_count.min() >= npts:
            break
        area_par *= fac
        count += 1
        if np.mod(count, 2) == 0:
            fac += 0.5
        if np.mod(count, max_iter) == 0:
            e = ("did not meet vertex requirement "
                 "after %d iterations" % max_iter)
            raise MeshLensCorrectionException(e)
    return t, area_par


def find_delaunay_with_max_vertices(bbox, nvertex):
    """optimize a delaunay triangulation of a PSLG to create an
    expected number of vertices

    Parameters
    ----------
    bbox : dict
        dictionary with keys vertices and segments representing a PSLG
    nvertex : int
        number of vertices for the triangulation to target

    Returns
    -------
    mesh : scipy.spatial.Delaunay
        resultant triangulation
    a : float
        area constraint used in the optimized triangulation
    """
    # find bracketing values
    a1 = a2 = 1e6
    t1 = calculate_mesh(a1, bbox, nvertex)
    afac = np.power(10., -np.sign(t1))
    while (
            np.sign(t1) ==
            np.sign(calculate_mesh(a2, bbox, nvertex))
          ):
        a2 *= afac
    val_at_root = -1
    nvtweak = nvertex
    while val_at_root < 0:
        a = scipy.optimize.brentq(
                calculate_mesh,
                a1,
                a2,
                args=(bbox, nvtweak, ))
        val_at_root = calculate_mesh(a, bbox, nvertex)
        a1 = a * 2
        a2 = a * 0.5
        nvtweak -= 1
    mesh = calculate_mesh(a, bbox, None, get_t=True)
    return mesh, a


def compute_barycentrics_legacy(coords, mesh):
    """legacy function to compute barycentric coordinates on mesh

    Parameters
    ----------
    coords : numpy.ndarray
        Nx2 array of points
    mesh : scipy.spatial.qhull.Delaunay
        triangular mesh

    Returns
    -------
    bcoords : numpy.ndarray
        Nx2 array of barycentric coordinates
    triangle_indices : numpy.ndarray
        simplex indices of barycentric coordinates
    """
    # https://en.wikipedia.org/wiki/Barycentric_coordinate_system#Conversion_between_barycentric_and_Cartesian_coordinates
    triangle_indices = mesh.find_simplex(coords)
    vt = np.vstack((
        np.transpose(mesh.points),
        np.ones(mesh.points.shape[0])))
    mt = np.vstack((np.transpose(coords), np.ones(coords.shape[0])))
    bary = np.zeros((3, coords.shape[0]))
    Rinv = []
    for tri in mesh.simplices:
        Rinv.append(np.linalg.inv(vt[:, tri]))
    for i in range(mesh.nsimplex):
        ind = np.argwhere(triangle_indices == i).flatten()
        bary[:, ind] = Rinv[i].dot(mt[:, ind])
    return np.transpose(bary), triangle_indices


def compute_barycentrics_native(coords, mesh):
    """convert coordinates to barycentric coordinates on mesh

    Parameters
    ----------
    coords : numpy.ndarray
        Nx2 array of points
    mesh : scipy.spatial.qhull.Delaunay
        triangular mesh

    Returns
    -------
    bcoords : numpy.ndarray
        Nx2 array of barycentric coordinates
    triangle_indices : numpy.ndarray
        simplex indices of barycentric coordinates
    """
    triangle_indices = mesh.find_simplex(coords)
    X = mesh.transform[triangle_indices, :2]
    Y = coords - mesh.transform[triangle_indices, 2]
    b = np.einsum('ijk,ik->ij', X, Y)
    bcoords = np.c_[b, 1 - b.sum(axis=1)]
    return bcoords, triangle_indices


def compute_barycentrics(coords, mesh, legacy_barycentrics=False, **kwargs):
    """convert coordinates to barycentric coordinates on mesh

    Parameters
    ----------
    coords : numpy.ndarray
        Nx2 array of points
    mesh : scipy.spatial.qhull.Delaunay
        triangular mesh
    legacy_barycentrics : boolean
        whether to use (slower) legacy method to find barycentrics.

    Returns
    -------
    bcoords : numpy.ndarray
        Nx2 array of barycentric coordinates
    triangle_indices : numpy.ndarray
        simplex indices of barycentric coordinates
    """
    if legacy_barycentrics:
        return compute_barycentrics_legacy(coords, mesh)
    else:
        return compute_barycentrics_native(coords, mesh)


def count_points_near_vertices(
        t, coords, bruteforce_simplex_counts=False,
        count_bincount=True, **kwargs):
    """enumerate coordinates closest to the vertices in a mesh

    Parameters
    ----------
    t : scipy.spatial.qhull.Delaunay
        triangular mesh
    coords : numpy.ndarray
        Nx2 array of points to assign to vertices on t
    bruteforce_simplex_counts : boolean
        whether to do a bruteforce simplex finding
    count_bincount : boolean
       use numpy.bincount based counting rather than legacy counting

    Returns
    -------
    pt_count : numpy.ndarray
        array with counts of points corresponding to indices of vertices in t
    """
    flat_tri = t.simplices.flatten()
    flat_ind = np.repeat(np.arange(t.nsimplex), 3)
    v_touches = []
    for i in range(t.npoints):
        v_touches.append(flat_ind[np.argwhere(flat_tri == i)])
    found = t.find_simplex(coords, bruteforce=bruteforce_simplex_counts)
    if count_bincount:
        bc = np.bincount(found, minlength=t.nsimplex)
        pt_count = np.array([
            bc[v_touches[i]].sum() for i in range(t.npoints)
        ])
    else:
        pt_count = np.zeros(t.npoints)
        for i in range(t.npoints):
            for j in v_touches[i]:
                pt_count[i] += np.count_nonzero(found == j)
    return pt_count


def create_regularization(ncols, ntiles, defaultL, transL, lensL):
    """create sparse regularization matrix with distinct
    translation and lens regularization factors

    Parameters
    ----------
    ncols : int
        columns to generate (based on solver matrix)
    ntiles : int
        number of tiles to regularize
    defaultL : float
        default regularization parameter lambda
    transL : float
        translation-only regularization factor applied to default
        regularization parameter
    lensL : float
        nonlinear-only regularization factor applied to default
        regularization parameter

    Returns
    -------
    rmat : :class:`scipy.sparse.csr`
        sparse regularization matrix
    """
    # regularizations: [affine matrix, translation, lens]
    reg = np.ones(ncols).astype('float64') * defaultL
    reg[0: (ntiles * 3)] *= transL
    reg[(ntiles * 3):] *= lensL
    rmat = sparse.eye(reg.size, dtype='float64', format='csr')
    rmat.data = reg
    return rmat


def create_thinplatespline_tf(
        mesh, solution,
        lens_dof_start,
        logger=default_logger,
        compute_affine=False):
    """create 2D Thin Plate Spline transform required to transform mesh to
    the solution derived from
    em_stitch.lens_correction.mesh_and_solve_transform.solve

    Parameters
    ----------
    mesh : scipy.spatial.qhull.Delaunay
        triangular source mesh object
    solution : list of numpy.ndarray
        list of numpy arrays of x and y vertex positions of the solution
    lens_dof_start : int
        start index describing degrees of freedom used in create_A
    logger : logging.Logger, optional
        logger used in this method
    compute_affine : bool
        Whether to compute an affine in Thin Plate Spline estimation.
        See renderapi.transform.ThinPlateSplineTransform

    Returns
    -------
    transform : renderapi.transform.ThinPlateSplineTransform
        transformation object deforming the mesh to the solution
    """

    dst = np.zeros_like(mesh.points)
    dst[:, 0] = mesh.points[:, 0] + solution[0][lens_dof_start:]
    dst[:, 1] = mesh.points[:, 1] + solution[1][lens_dof_start:]

    transform = renderapi.transform.ThinPlateSplineTransform()
    transform.estimate(mesh.points, dst, computeAffine=compute_affine)
    npts0 = transform.srcPts.shape[1]
    transform = transform.adaptive_mesh_estimate(max_iter=1000)
    npts1 = transform.srcPts.shape[1]

    logger.info(
            "adaptive_mesh_estimate reduced control points from %d to %d" %
            (npts0, npts1))

    transform.transformId = (
            datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")[:-3])
    transform.labels = None

    return transform


def new_specs_with_tf(ref_transform, tilespecs, transforms):
    """create a copy of each tilespec in tilespecs with the first transform
    being a reference to ref_transform and the second based on transforms
    provided in transforms.
    This likely expects a single transformation in the input tilespecs.

    Parameters
    ----------
    ref_transform : renderapi.transform.Transform
        transform to attach as a reference transform at index 0 for
        each tilespec
    tilespecs : list of renderapi.tilespec.TileSpec
        tilespecs to which ref_transform and the corresponding tranform
        from transforms should be applied
    transforms : list of renderapi.transform.Transform
        list of transforms of same length as tilespecs to apply as transform\
        at index 1 of each tilespec

    Returns
    -------
    newspecs : list of renderapi.tilespec.TileSpec
        copied tilespecs with transformations applied
    """
    newspecs = []
    for i in range(len(tilespecs)):
        newspecs.append(copy.deepcopy(tilespecs[i]))
        newspecs[-1].tforms.insert(0,
                                   renderapi.transform.ReferenceTransform(
                                    refId=ref_transform.transformId))
        newspecs[-1].tforms[1] = transforms[i]
    return newspecs


def solve(A, weights, reg, x0, b, precomputed_ATW=None, precomputed_ATWA=None,
          precomputed_K_factorized=None):
    """regularized weighted solve

    Parameters
    ----------
    A : :class:`scipy.sparse.csr`
        the matrix, N (equations) x M (degrees of freedom)
    weights : :class:`scipy.sparse.csr_matrix`
        N x N diagonal matrix containing weights
    reg : :class:`scipy.sparse.csr_matrix`
        M x M diagonal matrix containing regularizations
    x0 : :class:`numpy.ndarray`
        M x nsolve float constraint values for the DOFs
    b : :class:`numpy.ndarray`:
        N x nsolve float right-hand-side(s)
    precomputed_ATW : :class:`scipy.sparse.csc_matrix`
        value to use rather than computing A.T.dot(weights)
    precomputed_ATWA : :class:`scipy.sparse.csc_matrix`
        value to use rather than computing A.T.dot(weights).dot(A)
    precomputed_K_factorized : func
        factorized solve function to use rather than computing
        scipy.sparse.linalg.factorized(A.T.dot(weights).dot(A) + reg)

    Returns
    -------
    solution : list of numpy.ndarray
        list of numpy arrays of x and y vertex positions of solution
    errx : numpy.ndarray
        numpy array of x residuals
    erry : numpy.ndarray
        numpy array of y residuals

    """
    ATW = (A.transpose().dot(weights)
           if precomputed_ATW is None else precomputed_ATW)
    if precomputed_K_factorized is None:
        K = (ATW.dot(A) if precomputed_ATWA is None
             else precomputed_ATWA) + reg
        K_factorized = factorized(K)
    else:
        K_factorized = precomputed_K_factorized
    solution = []

    i = 0
    for x in x0:
        Lm = reg.dot(x) + ATW.dot(b[:, i])
        i += 1
        solution.append(K_factorized(Lm))

    errx = A.dot(solution[0]) - b[:, 0]
    erry = A.dot(solution[1]) - b[:, 1]

    return solution, errx, erry


def report_solution(errx, erry, transforms, criteria):
    """compile results, statistics, and messages for reporting information
    about lens correction solves

    Parameters
    ----------
    errx : numpy.ndarray
        numpy array of x residuals
    erry : numpy.ndarray
        numpy array of y residuals
    transforms : list of renderapi.transform.Transform
        list of transforms considered for the translation parameter
    criteria : dict
        criteria for good solve (deprecated)

    Returns
    -------
    translation : numpy.ndarray
        numpy array describing the resultant translations of the solution
    jresult : dict
        solution statistics dictionary
    message : str
        string reporting results
    """
    translation = np.array([tf.translation for tf in transforms])

    jresult = {}
    jresult['x_res_min'] = errx.min()
    jresult['x_res_max'] = errx.max()
    jresult['x_res_mean'] = errx.mean()
    jresult['x_res_std'] = errx.std()
    jresult['y_res_min'] = erry.min()
    jresult['y_res_max'] = erry.max()
    jresult['y_res_mean'] = erry.mean()
    jresult['y_res_std'] = erry.std()

    for k in jresult.keys():
        jresult[k] = np.round(jresult[k], 3)

    message = 'lens solver results [px]'
    for val in ['res']:
        for xy in ['x', 'y']:
            d = '\n%8s' % (xy + '_' + val + ' ')
            v = ''
            for calc in ['min', 'max', 'mean', 'std']:
                d += calc + ','
                k = xy + '_' + val + '_' + calc
                v += '%8.2f' % jresult[k]
            message += d + v

    return translation, jresult, message


def create_x0(nrows, tilespecs):
    """create initialization array x0

    Parameters
    ----------
    nrows : int
        number of rows in array (defined to match A[1])
    tilespecs : list of renderapi.tilespecs.TileSpec
        tilespecs from which initialization is built

    Returns
    -------
    x0 : numpy.ndarray
        initialization array x0
    """
    ntiles = len(tilespecs)
    x0 = []
    x0.append(np.zeros(nrows).astype('float64'))
    x0.append(np.zeros(nrows).astype('float64'))
    x0[0][0:ntiles] = np.zeros(ntiles)
    x0[1][0:ntiles] = np.zeros(ntiles)
    for i in range(ntiles):
        x0[0][i] = tilespecs[i].tforms[0].B0
        x0[1][i] = tilespecs[i].tforms[0].B1
    return x0


def create_A(matches, tilespecs, mesh, **kwargs):
    """create A matrix describing translation and lens correction

    Parameters
    ----------
    matches : list of dict
        list of match dictionaries in render format
    tilespecs : list of renderapi.tilespecs.TileSpec
        tilespecs to include in solve
    mesh : scipy.spatial.qhull.Delaunay
        mesh of input points as produced by
        em_stitch.lens_Correction.mesh_and_solve_transform._create_mesh

    Returns
    -------
    A : :class:`scipy.sparse.csr`
         matrix, N (equations) x M (degrees of freedom)
    wts : :class:`scipy.sparse.csr_matrix`
        N x N diagonal matrix containing weights
    b : :class:`numpy.ndarray`:
        N x nsolve float right-hand-side(s)
    lens_dof_start : int
        start index defined by degrees of freedom used to generate A
    """
    # let's assume translation halfsize
    dof_per_tile = 1
    dof_per_vertex = 1
    vertex_per_patch = 3
    nnz_per_row = 2*(dof_per_tile + vertex_per_patch * dof_per_vertex)
    nrows = sum([len(m['matches']['p'][0]) for m in matches])
    nd = nnz_per_row*nrows
    lens_dof_start = dof_per_tile*len(tilespecs)

    data = np.zeros(nd).astype('float64')
    b = np.zeros((nrows, 2)).astype('float64')
    indices = np.zeros(nd).astype('int64')
    indptr = np.zeros(nrows+1).astype('int64')
    indptr[1:] = np.arange(1, nrows+1)*nnz_per_row
    weights = np.ones(nrows).astype('float64')

    unique_ids = np.array(
            [t.tileId for t in tilespecs])

    # nothing fancy here, row-by-row
    offset = 0
    rows = 0

    for mi in range(len(matches)):
        m = matches[mi]
        pindex = np.argwhere(unique_ids == m['pId'])
        qindex = np.argwhere(unique_ids == m['qId'])

        npoint_pairs = len(m['matches']['q'][0])
        # get barycentric coordinates ready
        pcoords = np.transpose(
                np.vstack(
                    (m['matches']['p'][0],
                     m['matches']['p'][1])
                    )).astype('float64')
        qcoords = np.transpose(
                np.vstack(
                    (m['matches']['q'][0],
                     m['matches']['q'][1])
                    )).astype('float64')

        b[rows: (rows + pcoords.shape[0])] = qcoords - pcoords
        rows += pcoords.shape[0]
        pbary = compute_barycentrics(pcoords, mesh, **kwargs)
        qbary = compute_barycentrics(qcoords, mesh, **kwargs)

        mstep = np.arange(npoint_pairs) * nnz_per_row + offset

        data[mstep + 0] = 1.0
        data[mstep + 1] = -1.0
        data[mstep + 2] = pbary[0][:, 0]
        data[mstep + 3] = pbary[0][:, 1]
        data[mstep + 4] = pbary[0][:, 2]
        data[mstep + 5] = -qbary[0][:, 0]
        data[mstep + 6] = -qbary[0][:, 1]
        data[mstep + 7] = -qbary[0][:, 2]

        indices[mstep + 0] = pindex
        indices[mstep + 1] = qindex
        indices[mstep + 2] = (lens_dof_start +
                              mesh.simplices[pbary[1][:]][:, 0])
        indices[mstep + 3] = (lens_dof_start +
                              mesh.simplices[pbary[1][:]][:, 1])
        indices[mstep + 4] = (lens_dof_start +
                              mesh.simplices[pbary[1][:]][:, 2])
        indices[mstep + 5] = (lens_dof_start +
                              mesh.simplices[qbary[1][:]][:, 0])
        indices[mstep + 6] = (lens_dof_start +
                              mesh.simplices[qbary[1][:]][:, 1])
        indices[mstep + 7] = (lens_dof_start +
                              mesh.simplices[qbary[1][:]][:, 2])

        offset += npoint_pairs*nnz_per_row

    A = csr_matrix((data, indices, indptr), dtype='float64')

    wts = sparse.eye(weights.size, format='csr', dtype='float64')
    wts.data = weights
    return A, wts, b, lens_dof_start


def create_transforms(ntiles, solution):
    """create translation transformations from a solution array

    Parameters
    ----------
    ntiles : int
        number of tiles represented in the solution
    solution : list of numpy.ndarray
        list of numpy arrays of x and y positions result of solve

    Returns
    -------
    rtransforms : list of renderapi.transform.AffineModel
        transforms described by solution
    """
    rtransforms = []
    for i in range(ntiles):
        rtransforms.append(renderapi.transform.AffineModel(
                           B0=solution[0][i],
                           B1=solution[1][i]))
    return rtransforms


def estimate_stage_affine(t0, t1):
    """estimate affine transformation between translation components
    of tiles in t0 and tiles in t1 to give overall stage affine transformation.

    Parameters
    ----------
    t0 : list of renderapi.tilespec.TileSpec
        source tilespecs (initial position)
    t1 : list of renderapi.tilespec.TileSpec
        destination tilespecs (post-solve)
    
    Returns
    -------
    aff : renderapi.transform.AffineModel
        2D affine transform representing stage transformation
    """
    src = np.array([t.tforms[0].translation for t in t0])
    dst = np.array([t.tforms[1].translation for t in t1])
    aff = renderapi.transform.AffineModel()
    aff.estimate(src, dst)
    return aff


def _create_mesh(resolvedtiles, matches, nvertex,
                 return_area_triangle_par=False, **kwargs):
    """create mesh with a given number of vertices based on example tiles
        and pointmatches

    Parameters
    ----------
    resolvedtiles : renderapi.resolvedtiles.ResolvedTiles
        resolvedtiles containing a tilespec with mask, width, and height
        properties to use as a template for the mesh
    matches : list of dict
        list of point correspondences in render pointmatch format
    nvertex : int
        number of vertices for mesh
    return_area_triangle_par : boolean
        whether to return the area parameter used to generate the
        triangular mesh

    Returns
    -------
    mesh : scipy.spatial.qhull.Delaunay
        triangular mesh
    area_triangle_par : float
        max area constraint used in generating mesh
    """

    remove_weighted_matches(matches, weight=0.0)

    tilespecs = resolvedtiles.tilespecs
    example_tspec = tilespecs[0]

    tile_width = example_tspec.width
    tile_height = example_tspec.height
    maskUrl = example_tspec.ip[0].maskUrl

    coords = condense_coords(matches)
    nc0 = coords.shape[0]
    coords = smooth_density(
        coords,
        tile_width,
        tile_height,
        10, **kwargs)

    nc1 = coords.shape[0]
    logger.info(
        "\n  smoothing point density reduced points from %d to %d" %
        (nc0, nc1))
    if coords.shape[0] == 0:
        raise MeshLensCorrectionException(
            "no point matches left after smoothing density, \
            probably some sparse areas of matching")

    # create PSLG
    bbox = create_PSLG(
            tile_width,
            tile_height,
            maskUrl)

    # find delaunay with max vertices
    mesh, area_triangle_par = find_delaunay_with_max_vertices(
        bbox, nvertex)

    # and enforce neighboring matches to vertices
    mesh, area_triangle_par = force_vertices_with_npoints(
        area_triangle_par, bbox, coords, 3, **kwargs)

    return ((mesh, area_triangle_par) if return_area_triangle_par else mesh)


def _solve_resolvedtiles(
        resolvedtiles, matches, nvertex, regularization_lambda,
        regularization_translation_factor, regularization_lens_lambda,
        good_solve_dict,
        logger=default_logger, **kwargs):
    """generate lens correction from resolvedtiles and pointmatches

    Parameters
    ----------
    resolvedtiles : renderapi.resolvedtiles.ResolvedTiles
        resolvedtiles object on which transformation will be computed
    matches : list of dict
         point correspondences to consider in render pointmatch format
    nvertex :
        number of vertices in mesh
    regularization_lambda :  float
        lambda value for affine regularization
    regularization_translation_factor :  float
        translation factor of regularization
    regularization_lens_lambda :  float
        lambda value for lens regularization
    good_solve_dict :
        dictionary to define when a solve fails
    logger : logging.Logger
        logger to use in reporting
    Returns
    -------
    resolved : renderapi.resolvedtiles.ResolvedTiles
        new resolvedtiles object with derived lens correction applied
    new_ref_transform : renderapi.transform.leaf.ThinPlateSplineTransform
        derived lens correction transform
    jresult : dict
        dictionary of solve information
    """

    # FIXME this is done twice -- think through
    tilespecs = resolvedtiles.tilespecs
    example_tspec = tilespecs[0]

    mesh = _create_mesh(resolvedtiles, matches, nvertex, **kwargs)

    nend = mesh.points.shape[0]

    # logger = logging.getLogger(self.__class__.__name__)
    logger.info(
        "\n  aimed for %d mesh points, got %d" %
        (nvertex, nend))

    if mesh.points.shape[0] < 0.5*nvertex:
        raise MeshLensCorrectionException(
                "mesh coarser than intended")

    # prepare the linear algebra and solve
    A, weights, b, lens_dof_start = create_A(
        matches, tilespecs, mesh)

    x0 = create_x0(
        A.shape[1], tilespecs)

    reg = create_regularization(
        A.shape[1],
        len(tilespecs),
        regularization_lambda,
        regularization_translation_factor,
        regularization_lens_lambda)

    solution, errx, erry = solve(
        A, weights, reg, x0, b)

    transforms = create_transforms(
        len(tilespecs), solution)

    tf_trans, jresult, solve_message = report_solution(
            errx, erry, transforms, good_solve_dict)

    logger.info(solve_message)

    # check quality of solution
    if not all([
            errx.mean() < good_solve_dict['error_mean'],
            erry.mean() < good_solve_dict['error_mean'],
            errx.std() < good_solve_dict['error_std'],
            erry.std() < good_solve_dict['error_std']]):
        raise MeshLensCorrectionException(
                "Solve not good: %s" % solve_message)

    logger.debug(solve_message)

    new_ref_transform = create_thinplatespline_tf(
        mesh, solution, lens_dof_start, logger)

    bbox = example_tspec.bbox_transformed(tf_limit=0)
    tbbox = new_ref_transform.tform(bbox)
    bstr = 'new transform corners:\n'
    for i in range(bbox.shape[0]-1):
        bstr += "  (%0.1f, %0.1f) -> (%0.1f, %0.1f)\n" % (
                bbox[i, 0], bbox[i, 1],
                tbbox[i, 0], tbbox[i, 1])
        logger.info(bstr)

    new_tilespecs = new_specs_with_tf(
        new_ref_transform, tilespecs, transforms)

    stage_affine = estimate_stage_affine(tilespecs, new_tilespecs)
    sastr = (
        "affine estimate of tile translations:\n" +
        "  scale: {}\n".format(stage_affine.scale) +
        "  translation: {}\n".format(stage_affine.translation) +
        "  shear: {}\n".format(stage_affine.shear) +
        "  rotation: {}\n".format(np.degrees(stage_affine.rotation)))
    logger.info(sastr)

    resolved = renderapi.resolvedtiles.ResolvedTiles(
            tilespecs=new_tilespecs,
            transformList=[new_ref_transform])
    return resolved, new_ref_transform, jresult


class MeshAndSolveTransform(ArgSchemaParser):
    default_schema = MeshLensCorrectionSchema

    def solve_resolvedtiles_from_args(self):
        """use arguments to run lens correction

        Returns
        -------
        resolved : renderapi.resolvedtiles.ResolvedTiles
            new resolvedtiles object with derived lens correction applied
        new_ref_transform : renderapi.transform.leaf.ThinPlateSplineTransform
            derived lens correction transform
        jresult : dict
            dictionary of solve information
        """
        if 'tilespecs' in self.args:
            jspecs = self.args['tilespecs']
        else:
            jspecs = jsongz.load(self.args['tilespec_file'])

        self.tilespecs = np.array([
                renderapi.tilespec.TileSpec(json=j) for j in jspecs])

        if 'matches' in self.args:
            self.matches = self.args['matches']
        else:
            self.matches = jsongz.load(self.args['match_file'])

        return _solve_resolvedtiles(
            renderapi.resolvedtiles.ResolvedTiles(
                tilespecs=self.tilespecs, transformList=[]),
            self.matches, self.args["nvertex"],
            self.args["regularization"]["default_lambda"],
            self.args["regularization"]["translation_factor"],
            self.args["regularization"]["lens_lambda"],
            self.args["good_solve"],
            logger=self.logger
            )

    def run(self):
        self.resolved, self.new_ref_transform, jresult = (
            self.solve_resolvedtiles_from_args())

        new_path = None
        if 'outfile' in self.args:
            fname = self.args['outfile']
            if self.args['timestamp']:
                spf = fname.split(os.extsep, 1)
                spf[0] += '_%s' % self.new_ref_transform.transformId
                fname = os.extsep.join(spf)
            new_path = jsongz.dump(
                    self.resolved.to_dict(),
                    os.path.join(
                        self.args['output_dir'],
                        fname),
                    compress=self.args['compress_output'])
            new_path = os.path.abspath(new_path)

        fname = 'output.json'
        if self.args['timestamp']:
            fname = 'output_%s.json' % self.new_ref_transform.transformId

        self.args['output_json'] = os.path.join(
                self.args['output_dir'],
                fname)

        jresult['resolved_tiles'] = new_path

        self.output(jresult, indent=2)

        self.logger.info(" wrote solved tilespecs:\n  %s" % new_path)

        return



import logging
import os

import subprocess
import sys

from argschema import ArgSchemaParser

from .schemas import SetPermissionsSchema

logger = logging.getLogger(__name__)

example = {
        "client_mount_or_map": "/data/em-131fs3",  # could also be "Q:" e.g
        "fdir": 'lctest/T6.2019.04.18.110-120/000116/0/',
        "dir_setting": '777',
        'file_exts': ['.json', 'json.gz'],
        'file_setting': '777',
        "log_level": "INFO"
        }


def run_cmd(cmd):
    try:
        retcode = subprocess.call(cmd, shell=True)
        if retcode < 0:
            logger.info("command <%s> was terminated "
                        "by signal %d" % (cmd, -retcode))
        else:
            logger.info("command <%s> returned %d" % (cmd, retcode))
    except OSError as e:
        print("Execution failed:", e, file=sys.stderr)


def set_dirs(mydir, setting):
    if not setting:
        return
    cmd = 'sudo find %s -type d -exec chmod %s {} +' % (mydir, setting)
    run_cmd(cmd)


def set_files(mydir, exts, setting):
    if not setting:
        return
    cmd = 'sudo find %s -type f \( ' % mydir
    for i, ext in enumerate(exts):
        if i != 0:
            cmd += '-o '
        cmd += '-iname \*%s ' % ext
    cmd += '\) -exec chmod %s {} +' % setting
    run_cmd(cmd)


class SetPermissions(ArgSchemaParser):
    default_schema = SetPermissionsSchema

    def run(self):
        logger.setLevel(self.args['log_level'])

        mydir = os.path.join(
                self.args['client_mount_or_map'],
                self.args['fdir'])

        set_dirs(
                mydir,
                self.args['dir_setting'])

        set_files(
                mydir,
                self.args['file_exts'],
                self.args['file_setting'])


if __name__ == "__main__":
    p = SetPermissions(input_data=example)
    p.run()



from argschema import ArgSchema
from argschema.fields import (
    InputDir, InputFile, Str, List, Boolean, Int,
    OutputFile, Float)


class LensQuiverSchema(ArgSchema):
    transform_list = List(
        InputFile,
        required=True,
        description=("list of paths to transforms "
                     " or resolved tiles"))
    subplot_shape = List(
        Int,
        required=True,
        missing=[1, 1],
        default=[1, 1],
        description="sets the subplots for multiple plots")
    n_grid_pts = Int(
        required=True,
        missing=20,
        default=20,
        description="number of pts per axis for quiver grid")
    fignum = Int(
        required=True,
        missing=None,
        default=None,
        description="passed to plt.subplots to number the figure")
    arrow_scale = Float(
        required=True,
        missing=1.0,
        default=1.0,
        description="relative scale of arrows to axes")
    show = Boolean(
        required=True,
        missing=True,
        default=True,
        description=("show on screen?"))
    pdf_out = OutputFile(
        required=True,
        missing='./lens_corr_plots.pdf',
        default='./lens_corr_plots.pdf',
        description="where to write the pdf output")


class MontagePlotsSchema(ArgSchema):
    collection_path = InputFile(
        required=True,
        description="point matches from here")
    resolved_path = InputFile(
        required=True,
        description="resolved tiles from here")
    save_json_path = OutputFile(
        required=True,
        missing=None,
        default=None,
        description=("save residuals to this path if not None"))
    save_plot_path = OutputFile(
        required=True,
        missing=None,
        default=None,
        description=("save plot to this path if not None"))
    make_plot = Boolean(
        required=True,
        missing=True,
        default=True,
        description=("make a plot?"))
    show = Boolean(
        required=True,
        missing=True,
        default=True,
        description=("show on screen?"))
    pdf_out = OutputFile(
        required=True,
        missing=None,
        default=None,
        description="where to write the pdf output")


class ViewMatchesSchema(ArgSchema):
    collection_path = InputFile(
        required=False,
        description="if specified, will read collection from here")
    collection_basename = Str(
        required=True,
        missing="collection.json",
        default="collection.json",
        description=("basename for collection file if collection_path"
                     " not specified. will also check for .json.gz"))
    data_dir = InputDir(
        required=True,
        description=("directory containing image files. Will also be dir"
                     " dir for collection path, if not otherwise specified"))
    resolved_tiles = List(
        Str,
        required=True,
        missing=[
            "resolvedtiles.json.gz",
            "resolvedtiles_input.json.gz"],
        description=("will take the transform from the first file"
                     " matching this list, if possible"))
    transform_file = InputFile(
        required=False,
        description=("if provided, will get lens correction transform "
                     " from here"))
    view_all = Boolean(
        required=True,
        missing=False,
        default=False,
        description=("will plot all the pair matches. can be useful "
                     "for lens correction to file. probably not desirable "
                     "for montage"))
    show = Boolean(
        required=True,
        missing=True,
        default=True,
        description=("show on screen?"))
    match_index = Int(
        required=True,
        missing=0,
        default=0,
        description=("which index of self.matches to plot"))
    pdf_out = OutputFile(
        required=True,
        missing='./view_matches_output.pdf',
        default='./view_matches_output.pdf',
        description="where to write the pdf output")



import warnings

from marshmallow.warnings import ChangedInMarshmallow3Warning

from argschema import ArgSchema
from argschema.fields import (
    InputDir, InputFile, Float,
    Int, OutputFile, Str, Boolean)

warnings.simplefilter(
        action='ignore',
        category=ChangedInMarshmallow3Warning)


class GenerateEMTileSpecsParameters(ArgSchema):
    metafile = InputFile(
        required=True,
        description="metadata file containing TEMCA acquisition data")
    maskUrl = InputFile(
        required=False,
        default=None,
        missing=None,
        description="absolute path to image mask to apply")
    image_directory = InputDir(
        required=False,
        description=("directory used in determining absolute paths to images. "
                     "Defaults to parent directory containing metafile "
                     "if omitted."))
    maximum_intensity = Int(
        required=False, default=255,
        description=("intensity value to interpret as white"))
    minimum_intensity = Int(
        required=False, default=0,
        description=("intensity value to interpret as black"))
    z = Float(
        required=False,
        default=0,
        description=("z value"))
    sectionId = Str(
        required=False,
        description=("sectionId to apply to tiles during ingest.  "
                     "If unspecified will default to a string "
                     "representation of the float value of z_index."))
    output_path = OutputFile(
        required=False,
        description="directory for output files")
    compress_output = Boolean(
        required=False,
        missing=True,
        default=True,
        escription=("tilespecs will be .json or .json.gz"))



import warnings

from marshmallow.warnings import ChangedInMarshmallow3Warning

from argschema import ArgSchema
from argschema.schemas import DefaultSchema
from argschema.fields import (
    Nested, Int, Str, Boolean, List)

warnings.simplefilter(
        action='ignore',
        category=ChangedInMarshmallow3Warning)


class RenderClientParameters(DefaultSchema):
    host = Str(
        required=True, description='render host')
    port = Int(
        required=True, description='render post integer')
    owner = Str(
        required=True, description='render default owner')
    project = Str(
        required=True, description='render default project')
    client_scripts = Str(
        required=True, description='path to render client scripts')
    memGB = Str(
        required=False,
        default='5G',
        description='string describing java heap memory (default 5G)')
    validate_client = Boolean(
        required=False,
        default=False,
        description="will avoid problems on windows if we use use_rest")


class common_schema(ArgSchema):
    client_mount_or_map = Str(
        required=True,
        default="/data/em-131fs3",
        missing="/data/em-131fs3",
        description=("where the client sees the robocopied destination"
                     " windows example 'Q:'"))
    fdir = Str(
        required=True,
        description="appended to client_mount to find files")


class SetPermissionsSchema(common_schema):
    dir_setting = Str(
        required=True,
        default=None,
        missing=None,
        description=('setting to recursively apply to dirs. '
                     'robocopy writes at 755 and we want at least 775.'))
    file_exts = List(
        Str,
        required=True,
        missing=None,
        default=None,
        description="file extensions to change")
    file_setting = Str(
        required=True,
        default='777',
        missing='777',
        description=('setting to apply to files'
                     'robocopy writes at 744 and we want 777.'))


class resolved_schema(ArgSchema):
    resolved_file = Str(
        required=True,
        missing=None,
        description="basename of resolved_file")


class UpdateUrlSchema(common_schema, resolved_schema):
    backup_copy = Boolean(
        required=False,
        default=True,
        description="backup the resolved tilespecs file before overwriting")
    server_mount = Str(
        required=True,
        default="/data/em-131fs3",
        missing="/data/em-131fs3",
        description="where the render server sees the image files")
    image_directory = Str(
        required=False,
        missing=None,
        default=None,
        description=(" if missing, imageUrls are at server_mount/fdir/"
                     " if not missing, this."))


class UploadToRenderSchema(common_schema, resolved_schema):
    render = Nested(
        RenderClientParameters,
        required=True,
        description="parameters to connect to render server")
    stack = Str(
        required=False,
        description="name of destination stack in render")
    collection = Str(
        required=False,
        missing=None,
        description="name of destination collection in render")
    collection_file = Str(
        required=False,
        missing=None,
        description="collection file basename")
    close_stack = Boolean(
        required=False,
        default=True,
        missing=True,
        description="close stack or not after upload")


class SetUpdateUploadSchema(
        SetPermissionsSchema,
        UpdateUrlSchema,
        UploadToRenderSchema):
    pass



import warnings

from marshmallow.warnings import ChangedInMarshmallow3Warning
import marshmallow as mm

from argschema import ArgSchema
from argschema.schemas import DefaultSchema
from argschema.fields import (
    Boolean, InputDir, InputFile, Float, List,
    Int, OutputDir, Nested, Str, Dict)

warnings.simplefilter(
        action='ignore',
        category=ChangedInMarshmallow3Warning)


class regularization(DefaultSchema):
    default_lambda = Float(
        required=False,
        default=0.5,
        missing=0.5,
        description="regularization factor")
    translation_factor = Float(
        required=False,
        default=0.005,
        missing=0.005,
        description="transaltion factor")
    lens_lambda = Float(
        required=False,
        default=0.005,
        missing=0.005,
        description="regularization for lens parameters")


class good_solve_criteria(DefaultSchema):
    error_mean = Float(
        required=False,
        default=0.2,
        missing=0.2,
        description="maximum error mean [pixels]")
    error_std = Float(
        required=False,
        default=2.0,
        missing=2.0,
        description="maximum error std [pixels]")
    scale_dev = Float(
        required=False,
        default=0.1,
        missing=0.1,
        description="maximum allowed scale deviation from 1.0")


class MeshLensCorrectionSchema(ArgSchema):
    nvertex = Int(
        required=False,
        default=1000,
        missinf=1000,
        description="maximum number of vertices to attempt")
    tilespec_file = InputFile(
        required=False,
        description="path to json of tilespecs")
    tilespecs = List(
        Dict,
        required=False,
        description="list of dict of tilespecs")
    match_file = InputFile(
        required=False,
        description="path to json of matches")
    matches = List(
        Dict,
        required=False,
        description="list of dict of matches")
    regularization = Nested(regularization, missing={})
    good_solve = Nested(good_solve_criteria, missing={})
    output_dir = OutputDir(
        required=False,
        description="directory for output files")
    outfile = Str(
        required=False,
        description=("Basename to which resolved json output of "
                     "lens correction is written"))
    compress_output = Boolean(
        required=False,
        missing=True,
        default=True,
        description=("tilespecs will be .json or .json.gz"))
    timestamp = Boolean(
        required=False,
        missing=False,
        default=False,
        description="add a timestamp to basename output")

    @mm.post_load
    def one_of_two(self, data):
        for a, b in [
                ['tilespecs', 'tilespec_file'],
                ['matches', 'match_file']]:
            if (a in data) == (b in data):  # xor
                raise mm.ValidationError(
                        'must specify one and only one of %s or %s' % (a, b))


class LensCorrectionSchema(ArgSchema):
    data_dir = InputDir(
        required=True,
        description="directory containing metafile, images, and matches")
    output_dir = OutputDir(
        required=False,
        description="directory for output files")
    mask_file = InputFile(
        required=False,
        default=None,
        missing=None,
        description="mask to apply to each tile")
    nvertex = Int(
        required=False,
        default=1000,
        missinf=1000,
        description="maximum number of vertices to attempt")
    ransac_thresh = Float(
        required=False,
        default=5.0,
        missing=5.0,
        description="ransac outlier threshold")
    regularization = Nested(regularization, missing={})
    good_solve = Nested(good_solve_criteria, missing={})
    ignore_match_indices = List(
        Int,
        required=False,
        default=None,
        missing=None,
        description=("debug feature for ignoring certain indices"
                     " of the match collection"))
    compress_output = Boolean(
        required=False,
        missing=True,
        default=True,
        description=("tilespecs will be .json or .json.gz"))
    timestamp = Boolean(
        required=False,
        missing=False,
        default=False,
        description="add a timestamp to basename output")



from em_stitch.utils.utils import src_from_xy
from em_stitch.lens_correction.utils import (
        maps_from_tform,
        split_inverse_tform,
        remove_weighted_matches,
        estimate_stage_affine)
import renderapi
import numpy as np
import copy


def test_remove_weighted():
    npts = 100
    match = {
            'pId': 'a',
            'qId': 'b',
            'pGroupId': 'c',
            'qGroupId': 'd',
            'matches': {
                'p': (np.random.rand(2, npts) * 1000).tolist(),
                'q': (np.random.rand(2, npts) * 1000).tolist(),
                'w': np.random.randint(0, 2, npts).astype('float').tolist()
                }}
    orig = copy.deepcopy(match)
    remove_weighted_matches([match])
    assert np.all(np.isclose(match['matches']['w'], 1.0))
    keep = np.argwhere(
            np.isclose(np.array(orig['matches']['w']), 1.0)).flatten()
    assert keep.size == len(match['matches']['w'])

    for pq in ['p', 'q']:
        pqorig = np.array(orig['matches'][pq])[:, keep]
        pqfilt = np.array(match['matches'][pq])
        assert np.all(pqorig == pqfilt)


def test_estimate_affine():
    t0locs = np.random.rand(3, 9) * 1000
    t0locs[2, :] = 1.0
    M = np.array([
        [1.1, 0.02, 134],
        [-0.01, 1.02, -434],
        [0.0, 0.0, 1.0]])
    t1locs = M.dot(t0locs)
    t0specs = []
    t1specs = []
    for i in range(t0locs.shape[1]):
        t0specs.append(renderapi.tilespec.TileSpec(
                tforms=[
                    renderapi.transform.AffineModel(
                        B0=t0locs[0, i],
                        B1=t0locs[1, i])]))
        # t1specs have a lens correction transform
        t1specs.append(renderapi.tilespec.TileSpec(
                tforms=[
                    renderapi.transform.AffineModel(),
                    renderapi.transform.AffineModel(
                        B0=t1locs[0, i],
                        B1=t1locs[1, i])]))
    e = estimate_stage_affine(t0specs, t1specs)
    assert np.all(np.isclose(e.M, M))


def test_split_inverse():
    # the function just splits things up
    # so one doesn't run out of memory,
    # typically a problem with ThinPlateSpline
    tf = renderapi.transform.AffineModel()
    tf.M = np.array([
        [1.1, 0.02, 134],
        [-0.01, 1.02, -434],
        [0.0, 0.0, 1.0]])

    src = np.random.rand(1000000, 2) * 1000

    dst = tf.inverse_tform(src)
    dst_split_1 = split_inverse_tform(tf, src, block_size=2000000)
    assert np.all(dst_split_1 == dst)

    dst_split_2 = split_inverse_tform(tf, src, block_size=2000)
    assert np.all(dst_split_2 == dst)


def test_make_maps():
    # identity transform
    width = height = 1000
    tform = renderapi.transform.AffineModel()
    map1, map2, mask = maps_from_tform(
            tform, width, height)
    x = np.arange(width)
    y = np.arange(height)
    xt, yt = np.meshgrid(x, y)
    assert np.all(np.isclose(map1, xt))
    assert np.all(np.isclose(map2, yt))
    assert np.all(mask == 255)

    # something more interesting
    tform.M = np.array([
        [1.1, 0.02, 134],
        [-0.01, 1.02, -434],
        [0.0, 0.0, 1.0]])
    # the following looks easy but for ThinPlateSpline
    # and larger images, it blows up
    # so, the function we're testing does it better
    src = src_from_xy(x, y)
    dst = tform.inverse_tform(src)
    dx = dst[:, 0].reshape((y.size, x.size))
    dy = dst[:, 1].reshape((y.size, x.size))

    # check that the results match the easy way
    map1, map2, mask = maps_from_tform(
            tform, width, height)
    assert np.all(np.isclose(map1, dx))
    assert np.all(np.isclose(map2, dy))
    assert np.any(mask == 0)



from em_stitch.utils.utils import correction_grid
import json
import renderapi
import os
import glob
import numpy as np

test_files_dir = os.path.join(os.path.dirname(__file__), 'test_files')


def test_correction_grid():
    data_dir = os.path.join(test_files_dir, "montage_example")
    meta = glob.glob(os.path.join(
        data_dir, '_metadata*.json'))[0]
    with open(meta, 'r') as f:
        j = json.load(f)
    jtform = j[2]['sharedTransform']

    npts = 20
    src, dst = correction_grid(jtform, npts=npts)
    delta = dst - src
    mag = np.linalg.norm(delta, axis=1)
    assert mag.size == npts**2

    tf = renderapi.transform.ThinPlateSplineTransform(json=jtform)
    src, dst = correction_grid(tf, npts=npts)
    delta = dst - src
    mag = np.linalg.norm(delta, axis=1)
    assert mag.size == npts**2



from jinja2 import Environment, FileSystemLoader
import json
import pytest
import os
import copy
from bigfeta import jsongz
from em_stitch.lens_correction.lens_correction_solver import (
        LensCorrectionSolver, make_collection_json, one_file,
        LensCorrectionException, tilespec_input_from_metafile)
from em_stitch.lens_correction.mesh_and_solve_transform import \
        MeshAndSolveTransform
from em_stitch.utils.generate_EM_tilespecs_from_metafile import \
        GenerateEMTileSpecsModule
from tempfile import TemporaryDirectory
from marshmallow import ValidationError
import renderapi
import glob
import shutil

test_files_dir = os.path.join(os.path.dirname(__file__), 'test_files')
example_env = Environment(loader=FileSystemLoader(test_files_dir))


def json_template(env, template_file, **kwargs):
    template = env.get_template(template_file)
    d = json.loads(template.render(**kwargs))
    return d


@pytest.fixture(scope='module')
def solver_input_args():
    data_dir = os.path.join(test_files_dir, "lens_example")
    with TemporaryDirectory() as output_dir:
        yield json_template(
                example_env,
                "lens_solver_example.json",
                data_dir=data_dir,
                output_dir=output_dir)


@pytest.mark.parametrize('source', ['file', 'memory', 'fail'])
def test_solve_from_file_and_memory(solver_input_args, source):
    local_args = copy.deepcopy(solver_input_args)
    metafile = one_file(local_args['data_dir'], '_metadata*.json')
    templatefile = one_file(local_args['data_dir'], '_template*.json')
    compress = True
    with TemporaryDirectory() as output_dir:
        tspecin = tilespec_input_from_metafile(
                metafile,
                local_args['mask_file'],
                output_dir,
                local_args['log_level'],
                compress)
        gentspecs = GenerateEMTileSpecsModule(input_data=tspecin, args=[])
        gentspecs.run()
        cfile, counts = make_collection_json(
                templatefile,
                output_dir,
                compress,
                local_args['ransac_thresh'])

        solver_args = {
                'nvertex': local_args['nvertex'],
                'regularization': local_args['regularization'],
                'output_dir': output_dir,
                'compress_output': compress,
                'log_level': local_args['log_level']}

        if source == 'file':
            solver_args['tilespec_file'] = gentspecs.args['output_path']
            solver_args['match_file'] = cfile
            solver_args['outfile'] = 'resolvedtiles.json.gz'
            solver = MeshAndSolveTransform(input_data=solver_args, args=[])
            solver.run()
            with open(solver.args['output_json'], 'r') as f:
                tspec_path = json.load(f)['resolved_tiles']
            tfile = renderapi.resolvedtiles.ResolvedTiles(
                    json=jsongz.load(tspec_path))
            assert (
                    len(tfile.tilespecs) ==
                    len(gentspecs.tilespecs) ==
                    len(solver.resolved.tilespecs))
        if source == 'memory':
            solver_args['tilespecs'] = gentspecs.tilespecs
            solver_args['matches'] = jsongz.load(cfile)
            solver = MeshAndSolveTransform(input_data=solver_args, args=[])
            solver.run()
            assert (
                    len(gentspecs.tilespecs) ==
                    len(solver.resolved.tilespecs))
        if source == 'fail':
            solver_args['tilespec_file'] = gentspecs.args['output_path']
            solver_args['tilespecs'] = gentspecs.tilespecs
            solver_args['matches'] = jsongz.load(cfile)
            with pytest.raises(ValidationError):
                MeshAndSolveTransform(input_data=solver_args, args=[])
            solver_args.pop('tilespecs')
            solver_args['tilespec_file'] = gentspecs.args['output_path']
            solver_args['matches'] = jsongz.load(cfile)
            solver_args['match_file'] = cfile
            with pytest.raises(ValidationError):
                MeshAndSolveTransform(input_data=solver_args, args=[])


@pytest.mark.parametrize("timestamp", [True, False])
def test_solver(solver_input_args, timestamp):
    local_args = copy.deepcopy(solver_input_args)
    local_args['timestamp'] = timestamp
    with TemporaryDirectory() as output_dir:
        local_args['output_dir'] = output_dir
        lcs = LensCorrectionSolver(input_data=local_args, args=[])
        lcs.run()
        assert os.path.isfile(lcs.args['output_json'])
        with open(lcs.args['output_json'], 'r') as f:
            j = json.load(f)
        for f in j['output'].values():
            assert os.path.isfile(f)


def test_multifile_exception(solver_input_args):
    # this happened once, now this is here
    local_args = copy.deepcopy(solver_input_args)
    with TemporaryDirectory() as inout_dir:
        local_args['data_dir'] = inout_dir
        local_args['output_dir'] = inout_dir

        for f in glob.glob(
                os.path.join(
                    solver_input_args['data_dir'],
                    "*.json")):
            shutil.copy(
                    f,
                    os.path.join(
                        local_args['data_dir']))
        meta = glob.glob(
                os.path.join(
                    local_args['data_dir'], "_meta*.json"))[0]
        msp = os.path.splitext(
                os.path.basename(meta))
        meta2 = os.path.join(
                local_args['data_dir'],
                msp[0] + 'xxx' + '.' + msp[1])
        shutil.copy(meta, meta2)

        with pytest.raises(LensCorrectionException):
            lcs = LensCorrectionSolver(input_data=local_args, args=[])
            lcs.run()


@pytest.mark.parametrize("compress", [True, False])
def test_make_collection(solver_input_args, compress):
    ftemplate = glob.glob(
            os.path.join(
                solver_input_args['data_dir'],
                "_template_matches_*.json"))[0]
    with TemporaryDirectory() as output_dir:
        cfile, counts = make_collection_json(
                ftemplate,
                output_dir,
                compress,
                solver_input_args['ransac_thresh'])
        # for debug purposes, sometimes ignore some matches to see
        # how the solve does without them
        m = jsongz.load(cfile)
        n0 = len(m)
        cfile, counts = make_collection_json(
                ftemplate,
                output_dir,
                solver_input_args['ransac_thresh'],
                compress,
                ignore_match_indices=[0, 1])
        m = jsongz.load(cfile)
        n1 = len(m)
        assert n0 == (n1 + 2)



import time
import logging

import cv2
import numpy as np
from scipy import ndimage

import renderapi

from ..utils import utils as common_utils

logger = logging.getLogger(__name__)


def split_inverse_tform(tform, src, block_size):
    """
    Split and inverse transform the source array using the provided transformation.

    Parameters
    ----------
    tform : Any
        Transformation object.
    src : np.ndarray
        Source array.
    block_size : int
        Size of each block.

    Returns
    -------
    np.ndarray
        Inverse transformed array.
    """
    nsplit = np.ceil(float(src.shape[0]) / float(block_size))
    split_src = np.array_split(src, nsplit, axis=0)
    dst = []
    for s in split_src:
        dst.append(tform.inverse_tform(s))
    dst = np.concatenate(dst)
    return dst


def maps_from_tform(tform, width, height, block_size=10000, res=32):
    """
    Generate maps and a mask for remapping based on the provided transformation.

    Parameters
    ----------
    tform : Any
        Transformation object.
    width : int
        Width of the map.
    height : int
        Height of the map.
    block_size : int, optional
        Size of each block, by default 10000.
    res : int, optional
        cell resolution, by default 32.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray]
        Tuple containing map1, map2, and mask arrays.

    """
    t0 = time.time()

    x = np.arange(0, width + res, res)
    y = np.arange(0, height + res, res)
    src = common_utils.src_from_xy(x, y)
    idst = split_inverse_tform(tform, src, block_size)
    ix = idst[:, 0].reshape(y.size, x.size)
    iy = idst[:, 1].reshape(y.size, x.size)

    fx = np.arange(0, width)
    fy = np.arange(0, height)
    src = np.flipud(common_utils.src_from_xy(
        fx, fy, transpose=False).astype('float32'))
    src[0, :] *= (float(ix.shape[0] - 1) / y.max())
    src[1, :] *= (float(ix.shape[1] - 1) / x.max())

    dx = ndimage.map_coordinates(ix, src, order=1)
    dy = ndimage.map_coordinates(iy, src, order=1)

    map1 = dx.reshape((fy.size, fx.size)).astype('float32')
    map2 = dy.reshape((fy.size, fx.size)).astype('float32')

    # actually do it, to find a mask
    mask = np.ones_like(map1)
    mask = cv2.remap(mask, map1, map2, cv2.INTER_NEAREST)
    mask = np.uint8(mask * 255)
    t1 = time.time()
    logger.info(" created maps for remap:\n  took %0.1f seconds" % (t1 - t0))
    return map1, map2, mask


def estimate_stage_affine(t0, t1):
    """
    Estimate affine transformation between two sets of translations.

    Parameters
    ----------
    t0 : Iterable
        List of transformations (t0).
    t1 : Iterable
        List of transformations (t1).

    Returns
    -------
    renderapi.transform.AffineModel
        Estimated affine transformation.

    """
    src = np.array([t.tforms[0].translation for t in t0])
    dst = np.array([t.tforms[1].translation for t in t1])
    aff = renderapi.transform.AffineModel()
    aff.estimate(src, dst)
    return aff


def remove_weighted_matches(matches, weight=0.0):
    """
    Remove matches with specified weight.

    Parameters
    ----------
    matches : List[Dict[str, Any]]
        List of matches.
    weight : float, optional
        Weight threshold, by default 0.0.
    """
    for m in matches:
        ind = np.invert(np.isclose(np.array(m['matches']['w']), weight))
        m['matches']['p'] = np.array(m['matches']['p'])[:, ind].tolist()
        m['matches']['q'] = np.array(m['matches']['q'])[:, ind].tolist()
        m['matches']['w'] = np.array(m['matches']['w'])[ind].tolist()
    return



import json

import cv2
import numpy as np

import renderapi


def get_z_from_metafile(metafile):
    offsets = [
            {
              "load": "Tape147",
              "offset": 100000
            },
            {
              "load": "Tape148",
              "offset": 110000
            },
            {
              "load": "Tape148B",
              "offset": 110000
            },
            {
              "load": "Tape148A",
              "offset": 110000
            },
            {
              "load": "Tape149",
              "offset": 120000
            },
            {
              "load": "Tape151",
              "offset": 130000
            },
            {
              "load": "Tape162",
              "offset": 140000
            },
            {
              "load": "Tape127",
              "offset": 150000
            }]

    loads = np.array([i['load'] for i in offsets])

    with open(metafile, 'r') as f:
        j = json.load(f)
    try:
        tape = int(j[0]['metadata']['media_id'])
        offset = offsets[
                np.argwhere(loads == 'Tape%d' % tape).flatten()[0]]['offset']
    except (ValueError, IndexError):
        offset = 0
    grid = int(j[0]['metadata']['grid'])
    return offset + grid


def src_from_xy(x, y, transpose=True):
    xt, yt = np.meshgrid(x, y)
    src = np.vstack((xt.flatten(), yt.flatten())).astype('float')
    if not transpose:
        return src
    return src.transpose()


def correction_grid(transform, npts=20):
    """create src and dst for a thinplatespline transform
    Parameters
    ----------
    transform : renderapi.transform.ThinPlateSplineTransform
        or <>.to_dict() result of said transform

    Returns
    -------
    src : numpy array
        Nx2 where N = npts**2. linearly sampled grid
    dst : numpy array
        Nx2. The result of transform.tform(src)
    """

    if isinstance(transform, dict):
        transform = renderapi.transform.ThinPlateSplineTransform(
                json=transform)
    xymax = transform.srcPts.max(axis=1)
    src = src_from_xy(
            np.linspace(0, xymax[0], npts),
            np.linspace(0, xymax[1], npts))
    return src, transform.tform(src)


def pointmatch_filter(
        match, n_clusters=None, ransacReprojThreshold=10,
        n_cluster_pts=15, n_min_ignore=3, model='Affine'):
    """filter point matches via Similarity Model with
       local clustering, if specified.

    Parameters
    ----------
    match: dict
        pointmatch dict
    n_clusters: int
        number of clusters. If None, will be set by n_cluster_pts
    ransacReprojThreshold: float
        passed to cv2.estimateAffinePartial2D
    n_cluster_pts: int
        sets number of clusters by points. If n_cluster is specified,
        determines stopping criteria for finding clusters.
    n_min_ignore: int
        if the number of inliers <= this setting, all weights set to zero
        for the cluster

    Returns
    -------
    p: numpy array
        N x 2 array from match['matches']['p']
    q: numpy array
        N x 2 array from match['matches']['q']
    w: numpy array
        1.0/0.0 for inliers/outliers. Can be used in pointmatch
        dict match['matches']['w'] = w.tolist()
    labels: numpy array
        labels from kmeans clustering
    """

    p = np.array(match['matches']['p']).transpose().astype('float32')
    q = np.array(match['matches']['q']).transpose().astype('float32')

    # find some clusters with a minimum number of points
    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 100, 1.0)
    npts = len(match['matches']['w'])
    if n_clusters is None:
        n_clusters = npts // n_cluster_pts
    while True:
        _, labels, _ = cv2.kmeans(
                p,
                n_clusters,
                None,
                criteria,
                10,
                cv2.KMEANS_RANDOM_CENTERS)
        ulab, cnts = np.unique(labels, return_counts=True)
        if np.all(cnts >= n_cluster_pts) | (n_clusters == 1):
            break
        n_clusters -= 1

    labels = labels.flatten()

    if model == 'Affine':
        mfun = cv2.estimateAffine2D
    elif model == 'Similarity':
        mfun = cv2.estimateAffinePartial2D

    # run RANSAC on each cluster and track inliers
    w = np.zeros(npts)
    for u in ulab:
        cind = np.argwhere(labels == u).flatten()
        _, b = mfun(
                p[cind],
                q[cind],
                ransacReprojThreshold=ransacReprojThreshold)
        b = b.flatten()
        if np.count_nonzero(b) > n_min_ignore:
            w[cind[b != 0]] = 1.0

    return p, q, w, labels



import json
import os
import pathlib

import numpy

from argschema import ArgSchemaParser
from bigfeta import jsongz
import renderapi

from .schemas import GenerateEMTileSpecsParameters

# this is a modification of https://github.com/AllenInstitute/
# render-modules/blob/master/rendermodules/dataimport/
# generate_EM_tilespecs_from_metafile.py
# that does not depend on render-modules nor
# on a running render server


class RenderModuleException(Exception):
    pass


class GenerateEMTileSpecsModule(ArgSchemaParser):
    default_schema = GenerateEMTileSpecsParameters

    @staticmethod
    def image_coords_from_stage(stage_coords, resX, resY, rotation):
        cr = numpy.cos(rotation)
        sr = numpy.sin(rotation)
        x = stage_coords[0] / resX
        y = stage_coords[1] / resY
        return (int(x * cr + y * sr),
                int(-x * sr + y * cr))

    @staticmethod
    def tileId_from_basename(fname):
        return os.path.splitext(os.path.basename(fname))[0]

    @staticmethod
    def sectionId_from_z(z):
        return str(float(z))

    @staticmethod
    def ts_from_imgdata_tileId(imgdata, imgdir, x, y, tileId, 
                               minint=0, maxint=255, maskUrl=None,
                               width=3840, height=3840, z=None, sectionId=None,
                               scopeId=None, cameraId=None, pixelsize=None):
        raw_tforms = [renderapi.transform.AffineModel(B0=x, B1=y)]
        imageUrl = pathlib.Path(
            os.path.abspath(os.path.join(
                imgdir, imgdata['img_path']))).as_uri()
        if maskUrl is not None:
                maskUrl = pathlib.Path(maskUrl).as_uri()

        ip = renderapi.image_pyramid.ImagePyramid()
        ip[0] = renderapi.image_pyramid.MipMap(imageUrl=imageUrl,
                                               maskUrl=maskUrl)
        return renderapi.tilespec.TileSpec(
            tileId=tileId, z=z,
            width=width, height=height,
            minint=minint, maxint=maxint,
            tforms=raw_tforms,
            imagePyramid=ip,
            sectionId=sectionId, scopeId=scopeId, cameraId=cameraId,
            imageCol=imgdata['img_meta']['raster_pos'][0],
            imageRow=imgdata['img_meta']['raster_pos'][1],
            stageX=imgdata['img_meta']['stage_pos'][0],
            stageY=imgdata['img_meta']['stage_pos'][1],
            rotation=imgdata['img_meta']['angle'], pixelsize=pixelsize)

    def ts_from_imgdata(self, imgdata, imgdir, x, y,
                        minint=0, maxint=255, maskUrl=None,
                        width=3840, height=3840, z=None, sectionId=None,
                        scopeId=None, cameraId=None, pixelsize=None):
        tileId = self.tileId_from_basename(imgdata['img_path'])
        sectionId = (self.sectionId_from_z(z) if sectionId is None
                     else sectionId)
        return self.ts_from_imgdata_tileId(
            imgdata, imgdir, x, y, tileId, 
            minint, maxint, maskUrl,
            width, height, z, sectionId,
            scopeId, cameraId, pixelsize)

    @classmethod
    def ts_from_metadata(
            cls, md, image_directory, z, sectionId=None,
            minimum_intensity=0, maximum_intensity=255, maskUrl=None):
        roidata = md[0]['metadata']
        imgdata = md[1]['data']
        img_coords = {img['img_path']: cls.image_coords_from_stage(
            img['img_meta']['stage_pos'],
            img['img_meta']['pixel_size_x_move'],
            img['img_meta']['pixel_size_y_move'],
            numpy.radians(img['img_meta']['angle'])) for img in imgdata}

        minX, minY = numpy.min(numpy.array(list(img_coords.values())), axis=0)
        # assume isotropic pixels
        pixelsize = roidata['calibration']['highmag']['x_nm_per_pix']
        
        inputs = {
            "minint": minimum_intensity,
            "maxint": maximum_intensity,
            "z": z,
            "sectionId": sectionId,
            "maskUrl": maskUrl
        }

        tspecs = [
                cls.ts_from_imgdata_tileId(
                    img, image_directory,
                    img_coords[img['img_path']][0] - minX,
                    img_coords[img['img_path']][1] - minY,
                    cls.tileId_from_basename(img["img_path"]),
                    width=roidata['camera_info']['width'],
                    height=roidata['camera_info']['height'],
                    scopeId=roidata['temca_id'],
                    cameraId=roidata['camera_info']['camera_id'],
                    pixelsize=pixelsize, **inputs) for img in imgdata]
        return tspecs

    def run(self):
        with open(self.args['metafile'], 'r') as f:
            meta = json.load(f)
        roidata = meta[0]['metadata']
        imgdata = meta[1]['data']
        img_coords = {img['img_path']: self.image_coords_from_stage(
            img['img_meta']['stage_pos'],
            img['img_meta']['pixel_size_x_move'],
            img['img_meta']['pixel_size_y_move'],
            numpy.radians(img['img_meta']['angle'])) for img in imgdata}

        # if not imgdata:
        #     raise RenderModuleException(
        #         "No relevant image metadata found for metafile {}".format(
        #             self.args['metafile']))

        minX, minY = numpy.min(numpy.array(list(img_coords.values())), axis=0)
        # assume isotropic pixels
        pixelsize = roidata['calibration']['highmag']['x_nm_per_pix']

        imgdir = self.args.get(
            'image_directory', os.path.dirname(self.args['metafile']))

        self.render_tspecs = [
                self.ts_from_imgdata(
                    img, imgdir,
                    img_coords[img['img_path']][0] - minX,
                    img_coords[img['img_path']][1] - minY,
                    minint=self.args['minimum_intensity'],
                    maxint=self.args['maximum_intensity'],
                    width=roidata['camera_info']['width'],
                    height=roidata['camera_info']['height'],
                    z=self.args['z'],
                    sectionId=self.args.get('sectionId'),
                    scopeId=roidata['temca_id'],
                    cameraId=roidata['camera_info']['camera_id'],
                    pixelsize=pixelsize,
                    maskUrl=self.args['maskUrl']) for img in imgdata]

        if 'output_path' in self.args:
            self.args['output_path'] = jsongz.dump(
                self.tilespecs,
                self.args['output_path'],
                self.args['compress_output'])

    @property
    def tilespecs(self):
        tjs = [t.to_dict() for t in self.render_tspecs]
        return tjs


if __name__ == '__main__':
    gmod = GenerateEMTileSpecsModule()
    gmod.run()



from jinja2 import Environment, FileSystemLoader
import json
import pytest
import os
import copy
from em_stitch.montage.montage_solver import (
        MontageSolver, get_transform)
from tempfile import TemporaryDirectory
import glob
import shutil
from marshmallow import ValidationError

test_files_dir = os.path.join(os.path.dirname(__file__), 'test_files')
example_env = Environment(loader=FileSystemLoader(test_files_dir))


def json_template(env, template_file, **kwargs):
    template = env.get_template(template_file)
    d = json.loads(template.render(**kwargs))
    return d


@pytest.fixture(scope='module')
def solver_input_args():
    data_dir = os.path.join(test_files_dir, "montage_example")
    with TemporaryDirectory() as output_dir:
        yield json_template(
                example_env,
                "montage_solver_example.json",
                data_dir=data_dir,
                output_dir=output_dir,
                template_dir=test_files_dir)


def test_read_from(solver_input_args):
    meta = glob.glob(os.path.join(
        solver_input_args['data_dir'],
        '_metadata*.json'))[0]
    tf0 = get_transform(meta, '', {}, 'metafile')
    with TemporaryDirectory() as output_dir:
        tfp = os.path.join(output_dir, 'ref.json')
        with open(tfp, 'w') as f:
            json.dump(tf0.to_dict(), f)
        tf1 = get_transform('', tfp, {}, 'reffile')
        tf2 = get_transform('', '', tf0.to_dict(), 'dict')
        assert tf0 == tf1 == tf2


def test_solver(solver_input_args):
    local_args = copy.deepcopy(solver_input_args)
    with TemporaryDirectory() as output_dir:
        local_args['output_dir'] = output_dir
        ms = MontageSolver(input_data=local_args, args=[])
        ms.run()
        assert os.path.isfile(ms.args['output_json'])
        with open(ms.args['output_json'], 'r') as f:
            j = json.load(f)
        assert len(j) == 2
        for ij in j:
            assert os.path.isfile(
                    os.path.join(
                        ms.args['output_dir'],
                        ij['output']))
            assert os.path.isfile(
                    os.path.join(
                        ms.args['output_dir'],
                        ij['collection']))
            for k in ['x', 'y', 'mag']:
                assert ij[k]['mean'] < 2.0
                assert ij[k]['stdev'] < 2.0


def test_solver_no_output_dir(solver_input_args):
    local_args = copy.deepcopy(solver_input_args)
    with TemporaryDirectory() as output_dir:
        meta = glob.glob(os.path.join(
            local_args['data_dir'],
            '_metadata*.json'))[0]
        newmeta = os.path.join(
                output_dir,
                os.path.basename(meta))
        shutil.copy(meta, newmeta)
        local_args['data_dir'] = output_dir
        local_args.pop('output_dir')
        ms = MontageSolver(input_data=local_args, args=[])
        ms.run()
        assert os.path.isfile(ms.args['output_json'])
        with open(ms.args['output_json'], 'r') as f:
            j = json.load(f)
        assert len(j) == 2
        for ij in j:
            assert os.path.isfile(
                    os.path.join(
                        ms.args['output_dir'],
                        ij['output']))
            assert os.path.isfile(
                    os.path.join(
                        ms.args['output_dir'],
                        ij['collection']))
            for k in ['x', 'y', 'mag']:
                assert ij[k]['mean'] < 2.0
                assert ij[k]['stdev'] < 2.0


def test_solver_metafile_specify(solver_input_args):
    local_args = copy.deepcopy(solver_input_args)
    with TemporaryDirectory() as output_dir:
        local_args['output_dir'] = output_dir
        local_args['metafile'] = glob.glob(
                os.path.join(
                    local_args['data_dir'],
                    '_metadata*.json'))[0]
        local_args.pop('data_dir')
        ms = MontageSolver(input_data=local_args, args=[])
        ms.run()
        assert os.path.isfile(ms.args['output_json'])
        with open(ms.args['output_json'], 'r') as f:
            j = json.load(f)
        assert len(j) == 2
        for ij in j:
            assert os.path.isfile(
                    os.path.join(
                        ms.args['output_dir'],
                        ij['output']))
            assert os.path.isfile(
                    os.path.join(
                        ms.args['output_dir'],
                        ij['collection']))
            for k in ['x', 'y', 'mag']:
                assert ij[k]['mean'] < 2.0
                assert ij[k]['stdev'] < 2.0


def test_solver_schema_errors(solver_input_args):
    local_args = copy.deepcopy(solver_input_args)
    with TemporaryDirectory() as output_dir:
        local_args['output_dir'] = output_dir
        local_args['solver_templates'][0] = os.path.join(
                os.path.dirname(local_args['solver_templates'][0]),
                'file_does_not_exist.json')
        with pytest.raises(ValidationError):
            MontageSolver(input_data=local_args, args=[])

        local_args = copy.deepcopy(solver_input_args)
        local_args['output_dir'] = output_dir
        local_args.pop('data_dir')
        with pytest.raises(ValidationError):
            MontageSolver(input_data=local_args, args=[])



from em_stitch.utils.utils import pointmatch_filter, src_from_xy
import renderapi
import numpy as np
from shapely.geometry import Polygon, Point
from skimage.transform import PolynomialTransform as skPoly


def dummy_tilespec(width=1000, height=1000, transforms=[]):
    t = renderapi.tilespec.TileSpec()
    t.width = width
    t.height = height
    t.tforms = transforms
    return t


def random_affine(da=0.02, do=100):
    return renderapi.transform.AffineModel(
            M00=np.random.randn() * da + 1,
            M11=np.random.randn() * da + 1,
            M10=np.random.randn() * da,
            M01=np.random.randn() * da,
            B0=np.random.randn() * do,
            B1=np.random.randn() * do)


def random_polynomial(dparams=[100, 0.02, 0.00005]):
    params = np.array([
        [0, 1, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0]]).astype('float')
    params[:, 0] += np.random.randn(2) * dparams[0]
    params[:, 1:3] += np.random.randn(2, 2) * dparams[1]
    params[:, 3:] += np.random.randn(2, 3) * dparams[2]
    return renderapi.transform.Polynomial2DTransform(
            params=params)


def inverse(tspec, src):
    if isinstance(tspec.tforms[0], renderapi.transform.AffineModel):
        itf = tspec.tforms[0].invert()
    elif isinstance(
            tspec.tforms[0], renderapi.transform.Polynomial2DTransform):
        x = np.linspace(0, tspec.width, 10)
        y = np.linspace(0, tspec.height, 10)
        src = src_from_xy(x, y)
        dst = tspec.tforms[0].tform(src)
        # renderapi doesn't have a good estimator, use skimage
        invtf = skPoly()
        invtf.estimate(dst, src, order=2)
        itf = renderapi.transform.Polynomial2DTransform(
                params=invtf.params)
    return itf.tform(src).transpose().tolist()


def random_transform(tftype):
    if tftype == 'affine':
        return random_affine()
    elif tftype == 'polynomial':
        return random_polynomial()


def dummy_match(npts=100, tform_type='affine'):
    t0 = dummy_tilespec(transforms=[random_transform(tform_type)])
    t1 = dummy_tilespec(transforms=[random_transform(tform_type)])
    p0 = Polygon(t0.bbox_transformed(ndiv_inner=5))
    p1 = Polygon(t1.bbox_transformed(ndiv_inner=5))
    overlap = p0.intersection(p1)
    pts = []
    ow = overlap.bounds[2] - overlap.bounds[0]
    oh = overlap.bounds[3] - overlap.bounds[1]
    while len(pts) < npts:
        pt = Point(
                np.random.rand() * ow + overlap.bounds[0],
                np.random.rand() * oh + overlap.bounds[1])
        if overlap.contains(pt):
            pts.append(pt)
    src = np.array([np.array(p.coords[0]) for p in pts])
    match = {}
    for k in ['pId', 'qId', 'pGroupId', 'qGroupId']:
        match[k] = k
    match['matches'] = {
            'p': inverse(t0, src),
            'q': inverse(t1, src),
            'w': [1.0] * npts
            }
    return match


def test_one_cluster_filter():
    # check the size and shape of things
    m = dummy_match()
    p, q, w, labels = pointmatch_filter(m, n_clusters=1)
    unique_labels = np.unique(labels)
    assert np.all(p.shape == q.shape)
    assert p.shape[0] == w.size
    assert p.shape[0] == len(m['matches']['p'][0])
    assert q.shape[0] == len(m['matches']['q'][0])
    assert w.size == len(m['matches']['w'])
    assert unique_labels.size == 1
    assert np.all(np.isclose(w, 1.0))

    # make a small change to the match
    m['matches']['p'][0][12] += 3.0
    p, q, w, labels = pointmatch_filter(
            m, n_clusters=1, ransacReprojThreshold=10.0)
    assert np.all(np.isclose(w, 1.0))

    # make a larger change that should be filtered
    m['matches']['p'][0][12] += 20.0
    p, q, w, labels = pointmatch_filter(
            m, n_clusters=1, ransacReprojThreshold=10.0)
    assert np.count_nonzero(np.isclose(w, 0.0)) == 1

    # make a really small match, which will be filtered because of size
    m = dummy_match(npts=5)
    p, q, w, labels = pointmatch_filter(
            m, n_clusters=1, n_min_ignore=5)
    assert np.all(np.isclose(w, 0.0))

    # example is exact affine, so we can dial down the ransac parameter
    m = dummy_match()
    p, q, w, labels = pointmatch_filter(
            m, n_clusters=1, ransacReprojThreshold=0.1)
    assert np.all(np.isclose(w, 1.0))

    # but with Similarity instead of Affine, will not match all
    m = dummy_match(npts=100)
    p, q, w, labels = pointmatch_filter(
            m, n_clusters=1, model='Similarity', ransacReprojThreshold=0.1)
    assert np.any(np.isclose(w, 0.0))

    # and, with underlying non-linearity, affine won't match all either
    m = dummy_match(npts=100, tform_type='polynomial')
    p, q, w, labels = pointmatch_filter(
            m, n_clusters=1, model='Affine', ransacReprojThreshold=0.1)
    assert np.any(np.isclose(w, 0.0))


def test_n_cluster_filter():
    # start with underlying non-linearity
    npts = 100
    m = dummy_match(npts=npts, tform_type='polynomial')

    # in general, the number of inliers should increase
    # with more clusters. There is some randomness in RANSAC
    # so, we'll just check that the trend is such
    ncs = np.arange(1, 6)
    nin = np.zeros_like(ncs)
    for i in range(ncs.size):
        p, q, w, labels = pointmatch_filter(
                m,
                n_clusters=ncs[i],
                model='Affine',
                ransacReprojThreshold=10.0)
        nin[i] = np.count_nonzero(np.isclose(w, 1.0))
    lp = np.polyfit(ncs, nin, 1)
    assert lp[0] > 0

    # above was 100 pts divided into up to 5 clusters
    # or, about 20 pts per cluster
    # we can force it into a smaller number
    # n_cluster_pts
    npts = 100
    nc_ask = 5
    m = dummy_match(npts=npts, tform_type='polynomial')
    p, q, w, labels = pointmatch_filter(
            m,
            n_clusters=nc_ask,
            model='Affine',
            ransacReprojThreshold=10.0,
            n_cluster_pts=25)
    nc_get = np.unique(labels).size
    assert nc_get < nc_ask

    # or, we can set n_clusters=None and have the function decide
    m = dummy_match(npts=npts, tform_type='polynomial')
    p, q, w, labels = pointmatch_filter(
            m,
            n_clusters=None,
            model='Affine',
            ransacReprojThreshold=10.0,
            n_cluster_pts=25)
    nc_get = np.unique(labels).size
    assert nc_get < 5



import logging
import os

from argschema import ArgSchemaParser
from bigfeta import jsongz
import renderapi

from .schemas import UploadToRenderSchema

logger = logging.getLogger(__name__)

example = {
        "client_mount_or_map": "/data/em-131fs3",  # could also be "Q:" e.g
        "fdir": 'lctest/T6.2019.04.18.110-120/000116/0/',
        "resolved_file": "resolvedtiles_AffineModel_0.json.gz",
        "collection_file": "collection.json.gz",
        "render": {
                "host": "em-131fs",
                "port": 8987,
                "owner": "danielk",
                "project": "montage_test",
                "client_scripts": "/allen/aibs/pipeline/image_processing/volume_assembly/render-jars/production/scripts"
              },
        "stack": "T4_test",
        "collection": "T4_test",
        "log_level": "INFO"
        }


def upload_resolved_file(render_params, stack, resolved_file, close_stack):
    if resolved_file is None:
        return

    resolved = renderapi.resolvedtiles.ResolvedTiles(
            json=jsongz.load(resolved_file))

    render = renderapi.connect(**render_params)

    renderapi.stack.create_stack(stack, render=render)
    renderapi.client.import_tilespecs(
            stack,
            resolved.tilespecs,
            sharedTransforms=resolved.transforms,
            use_rest=True,
            render=render)
    if close_stack:
        renderapi.stack.set_stack_state(
                stack,
                state='COMPLETE',
                render=render)

    logger.info("imported %d tilespecs to render from %s" % (
        len(resolved.tilespecs), resolved_file))

    url = ("\nhttp://" +
           "%s:%d" % (render_params['host'], render_params['port']) +
           "/render-ws/view/stacks.html?ndvizHost=em-131fs%3A8001" +
           "&renderStack=%s" % stack +
           "&renderStackOwner=%s" % render_params['owner'] +
           "&renderStackProject=%s" % render_params['project'])
    logger.info(url)

    return


def upload_collection_file(render_params, collection, collection_file):
    if collection is None:
        return

    render = renderapi.connect(**render_params)

    matches = jsongz.load(collection_file)

    renderapi.pointmatch.import_matches(
            collection,
            matches,
            render=render)

    logger.info("imported %d pointmatches to render from %s" % (
        len(matches), collection_file))


class UploadToRender(ArgSchemaParser):
    default_schema = UploadToRenderSchema

    def run(self):
        logger.setLevel(self.args['log_level'])

        resolved_file = os.path.join(
                self.args['client_mount_or_map'],
                self.args['fdir'],
                self.args['resolved_file'])

        upload_resolved_file(
                self.args['render'],
                self.args['stack'],
                resolved_file,
                self.args['close_stack'])

        if self.args['collection_file']:
            collection_file = os.path.join(
                    self.args['client_mount_or_map'],
                    self.args['fdir'],
                    self.args['collection_file'])

            upload_collection_file(
                    self.args['render'],
                    self.args['collection'],
                    collection_file)


if __name__ == "__main__":
    umod = UploadToRender(input_data=example)
    umod.run()



#!/usr/bin/env python
from setuptools import setup,find_packages
import sys
from setuptools.command.test import test as TestCommand


class PyTest(TestCommand):
    user_options = [('pytest-args=', 'a', "Arguments to pass to pytest")]

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = ""

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import shlex
        import pytest
        self.pytest_args += " --cov=em_stitch --cov-report html "\
                            "--junitxml=test-reports/test.xml " \
                            "--ignore=src"

        errno = pytest.main(shlex.split(self.pytest_args))
        sys.exit(errno)


with open('test_requirements.txt', 'r') as f:
    test_required = f.read().splitlines()

with open('requirements.txt', 'r') as f:
    required = f.read().splitlines()

setup(name='em_stitch',
      use_scm_version=True,
      description='a python package for stitching EM images',
      author_email='danielk@alleninstitute.org',
      url='https://github.com/AllenInstitute/em_stitch',
      packages=find_packages(),
      setup_requires=['setuptools_scm'],
      install_requires=required,
      tests_require=test_required,
      cmdclass={'test': PyTest})



#!/usr/bin/env python
''' Converts metafiles to p->q collections.
    Jay Borseth 2019.02.15

'''

import argparse
from enum import IntEnum
import glob
import json
import os
import sys


# Position codes in metafile
class Edge(IntEnum):
    INVALID = 0
    CENTER = 1  # unused
    LEFT = 2
    TOP = 3
    RIGHT = 4


class MetaToCollection(object):
    ''' Converts a raw TEMCA metafile into a collection
        json file which the render stack can consume.
    '''

    def tile_from_raster_pos(self, args, col, row, direction=None):
        ''' returns a neighboring tile given a col, row.
            direction is either
            None (return this tile), LEFT, RIGHT, or TOP
            If the tile has no neighbor in the given direction,
            None is returned
        '''
        if direction is None:
            return args.raster_pos_lookup[str(col) + "_" + str(row)]
        elif direction == Edge.LEFT:
            if col > 0:
                try:
                    return args.raster_pos_lookup[
                            str(col - 1) + "_" + str(row)]
                except:
                    return None
            else:
                return None
        elif direction == Edge.RIGHT:
            if col < args.tcols:
                try:
                    return args.raster_pos_lookup[
                            str(col + 1) + "_" + str(row)]
                except:
                    return None
            else:
                return None
        elif direction == Edge.TOP:
            if row > 0:
                try:
                    return args.raster_pos_lookup[
                            str(col) + "_" + str(row - 1)]
                except:
                    return None
            else:
                return None

    def tile_from_tile(self, args, tile, direction=None):
        ''' returns a neighboring tile given a tile.
            direction is either None (return this tile), LEFT, RIGHT, or TOP
        '''
        col, row = tile['img_meta']['raster_pos']
        return self.tile_from_raster_pos(args, col, row, direction)

    def create_raster_pos_dict(self, args):
        ''' create the look up dictionary for raster pos to nodes '''
        args.raster_pos_lookup = {}
        for tile in args.data:
            rp = tile['img_meta']['raster_pos']
            col, row = rp
            args.raster_pos_lookup[str(col) + "_" + str(row)] = tile

    def get_meta_and_montage_files(self, rootdir):
        '''get the names of the meta and montage files'''
        for name in glob.glob(os.path.join(rootdir, r"_meta*.*")):
            meta = name
        montage = None
        for name in glob.glob(os.path.join(rootdir, r"_montage*.*")):
            montage = name
        return (meta, montage)

    def process(self, args):
        ''' the main thing.'''
        ''' read in the metadata file and extract relevant info'''
        rootdir = args.directory
        try:
            args.meta_file, args.montage_file = \
                    self.get_meta_and_montage_files(rootdir)

            with open(args.meta_file) as data_file:
                json_data = json.load(data_file)
        except:
            raise Exception("Cannot find or parse metafile in: " +
                            args.directory)

        metadata = args.metadata = json_data[0]['metadata']
        data = args.data = json_data[1]['data']

        temca_id = metadata["temca_id"]
        session_id = metadata["session_id"]
        grid = metadata["grid"]
        specimen_id = metadata["specimen_id"]
        if "tape_id" in metadata:
            tape_id = metadata["tape_id"]
        else:
            tape_id = None

        gid = (
                str(specimen_id) + '_' +
                str(temca_id) + '_' +
                str(tape_id) + '_' +
                str(session_id) + '_' +
                str(grid))
        qGroupId = pGroupId = gid

        # total number of rows and cols
        args.trows = max([tile['img_meta']['raster_pos'][1] for tile in data])
        args.tcols = max([tile['img_meta']['raster_pos'][0] for tile in data])
        print('rows: ', args.trows, ', cols: ', args.tcols)

        # create a dictionary to look up neighboring tiles
        self.create_raster_pos_dict(args)

        samples = []
        tilespecs = []

        # for all tiles
        for index, tile in enumerate(data):

            # tile == 'q' tile, where the template search is taking place

            qId = tile["img_path"]
            # hmm, munge the filenames?
            qId = qId.replace(".tif", "")
            tilespec = {
                'tileId': qId,
                'xstage': float(tile["img_meta"]['stage_pos'][0]),
                'ystage': float(tile["img_meta"]['stage_pos'][1])
            }
            tilespecs.append(tilespec)

            p = [[], []]
            q = [[], []]
            w = []

            if 'matcher' in tile:
                for match in tile['matcher']:
                    position = match['position']
                    match_quality = match['match_quality']
                    if match_quality == -1:
                        # -1 is a flag indicating no matches
                        # are possible for this tile edge
                        continue
                    neighbor = self.tile_from_tile(args, tile, position)
                    # neighbor == 'p' tile, which
                    # contains the original template
                    if neighbor:
                        p = [match["pX"], match["pY"]]
                        q = [match["qX"], match["qY"]]
                        w = [1] * len(match["pX"])
                        # hmm, munge the filenames?
                        pId = neighbor["img_path"]
                        pId = pId.replace(".tif", "")

                        samples.append({
                            'pId': pId,
                            'qId': qId,
                            'pGroupId': pGroupId,
                            'qGroupId': qGroupId,
                            'matches': {
                                'p': p,
                                'q': q,
                                'w': w,
                                'match_count': len(w),
                            }
                        })

        #with open(args.output_file, 'w') as f:
        #    json.dump(samples, f, indent=2)

        return samples


def main(args):
    parent_parser = argparse.ArgumentParser(
        description='Converts raw TEMCA metadata files to render collections')

    parent_parser.add_argument(
        'directory',
        help='the directory to process.',
        metavar="",
        nargs='?',
        default="/allen/programs/celltypes/workgroups/em-connectomics/danielk/lcdata/lens_correction16/000000/0")

    parent_parser.add_argument(
        '-o',
        '--output_file',
        type=str,
        default="test.json",
        metavar="",
        help='name of the json output file')

    args = parent_parser.parse_args(args)

    m2c = MetaToCollection()
    return m2c.process(args)


if __name__ == "__main__":
    main(sys.argv[1:])



import logging
import os
import pathlib
import shutil

from six.moves import urllib

from argschema import ArgSchemaParser
from bigfeta import jsongz
import renderapi

from .schemas import UpdateUrlSchema

logger = logging.getLogger(__name__)

example = {
        "backup_copy": True,
        "client_mount_or_map": "/data/em-131fs3",  # could also be "Q:" e.g
        "fdir": 'lctest/T6.2019.04.18.110-120/000116/0/',
        "resolved_file": "resolvedtiles_AffineModel_0.json.gz",
        "server_mount": "/data/em-131fs3",  # leave as posix
        "image_directory": None,
        "log_level": "INFO"
        }


def backup(f):
    for ext in ['.json', '.json.gz']:
        if f.endswith(ext):
            b = f.split(ext)[0]
            break
    i = 0
    newf = f
    while os.path.isfile(newf):
        # were having some permission errors
        # from multiple users writing this backup
        # file.
        newf = b + '_backup_%d' % i + ext
        i += 1
    shutil.copy(f, newf)
    logger.info("copied %s to %s" % (f, newf))


class UpdateUrls(ArgSchemaParser):
    default_schema = UpdateUrlSchema

    def run(self):
        logger.setLevel(self.args['log_level'])

        resolved_path = os.path.join(
                self.args['client_mount_or_map'],
                self.args['fdir'],
                self.args['resolved_file'])

        if self.args['backup_copy']:
            backup(resolved_path)

        resolved = renderapi.resolvedtiles.ResolvedTiles(
                json=jsongz.load(resolved_path))

        if self.args['image_directory'] is None:
            self.args['image_directory'] = pathlib.PurePosixPath(
                    self.args['server_mount'],
                    self.args['fdir'])

        for t in resolved.tilespecs:
            for k in ['imageUrl', 'maskUrl']:
                s = t.ip['0'][k]
                if s:
                    orig = urllib.parse.unquote(urllib.parse.urlparse(s).path)
                    t.ip['0'][k] = pathlib.PurePosixPath(
                            self.args['image_directory'],
                            os.path.basename(orig)).as_uri()

        self.args['resolved_file'] = jsongz.dump(
                resolved.to_dict(), resolved_path, compress=None)

        logger.info("updated tilespec urls in %s" % resolved_path)


if __name__ == '__main__':
    u = UpdateUrls(input_data=example)
    u.run()



import glob
import json
import logging
import os
import warnings

import cv2
import numpy as np

from argschema import ArgSchemaParser
from bigfeta import jsongz
import renderapi

from .schemas import LensCorrectionSchema
from ..utils.generate_EM_tilespecs_from_metafile import (
    GenerateEMTileSpecsModule)
from ..utils import utils as common_utils
from .mesh_and_solve_transform import MeshAndSolveTransform
from . import utils

warnings.simplefilter(action='ignore', category=FutureWarning)

logger = logging.getLogger(__name__)

example = {
        "data_dir": "/data/em-131fs3/lctest/T4_07/20190315142205_reference/0",
        "output_dir": "/data/em-131fs3/lctest/T4_07/20190315142205_reference/0",
        "mask_file": None,
        "ransac_thresh": 10,
        "nvertex": 1000,
        "regularization": {
            "default_lambda": 1.0,
            "lens_lambda": 1.0,
            "translation_factor": 1e-3
            },
        "log_level": "INFO",
        "write_pdf": False
        }


class LensCorrectionException(Exception):
    pass


def one_file(fdir, fstub):
    """
    Get a single file matching a directory and filename pattern.

    Parameters
    ----------
    fdir : str
        The directory where the file should be located.
    fstub : str
        The filename pattern to match.

    Returns
    -------
    str
        The full path of the single file matching the pattern.

    Raises
    ------
    LensCorrectionException
        If no file or more than one file is found matching the pattern.
    """
    fullstub = os.path.join(fdir, fstub)
    files = glob.glob(fullstub)
    lf = len(files)
    if lf != 1:
        raise LensCorrectionException(
            "expected 1 file, found %d for %s" % (
                lf,
                fullstub))
    return files[0]


def tilespec_input_from_metafile(
        metafile, mask_file, output_dir, log_level, compress):
    """
    get tilespec input data from a metafile.

    Parameters
    ----------
    metafile : str
        Path to the metafile.
    mask_file : str
        Path to the mask file.
    output_dir : str
        Directory where the output will be stored.
    log_level : int
        Log level for the operation.
    compress : bool
        Whether to compress the output.

    Returns
    -------
    Dict[str, Union[str, int, bool]]
        A dictionary containing the generated tilespec input data.
    """
    result = {}
    result['metafile'] = metafile

    with open(metafile, 'r') as f:
        jmeta = json.load(f)

    result['z'] = np.random.randint(0, 1000)
    result['minimum_intensity'] = 0
    bpp = jmeta[0]['metadata']['camera_info']['camera_bpp']
    result['maximum_intensity'] = int(np.power(2, bpp) - 1)
    result['sectionId'] = jmeta[0]['metadata']['grid']
    result['maskUrl'] = mask_file
    result['output_path'] = os.path.join(output_dir, 'raw_tilespecs.json')
    result['log_level'] = log_level
    result['compress_output'] = compress
    return result


def filter_match_collection(
        matches, threshold, model="Similarity",
        n_clusters=None, n_cluster_pts=20, ransacReprojThreshold=40.,
        ignore_match_indices=(),
        input_n_key="n_from_gpu", output_n_key="n_after_filter"):
    """
    Filter a collection of matches based on specified criteria.

    Parameters
    ----------
    matches : List[Dict[str, Any]]
        The collection of matches to filter.
    threshold : float
        Threshold value.
    model : str, optional
        Model type, by default "Similarity".
    n_clusters : int, optional
        Number of clusters, by default None.
    n_cluster_pts : int, optional
        Number of cluster points, by default 20.
    ransacReprojThreshold : float, optional
        RANSAC reprojection threshold, by default 40.0.
    ignore_match_indices : Optional[Iterable[int]], optional
        Indices of matches to ignore, by default None.
    input_n_key : str, optional
        Key to store input count in counts dictionary, by default "n_from_gpu".
    output_n_key : str, optional
        Key to store output count in counts dictionary, by default "n_after_filter".

    Returns
    -------
    Tuple[List[Dict[str, Any]], List[Dict[str, int]]]
        A tuple containing the filtered matches and their corresponding counts.
    """
    ignore_match_indices = (set() if ignore_match_indices is None else ignore_match_indices)
    ignore_match_indices = set(ignore_match_indices)

    counts = []
    new_matches = []

    for i, m in enumerate(matches):
        input_n = len(m["matches"]["p"][0])

        _, _, w, _ = common_utils.pointmatch_filter(
            m,
            n_clusters=n_clusters,
            n_cluster_pts=n_cluster_pts,

            )

        m["matches"]["w"] = w.tolist()

        output_n = np.count_nonzero(w)

        counts.append({
            input_n_key: input_n,
            output_n_key: output_n})

        # original version ran pointmatch filtering for these and then filtered
        #   them out of the matches.  Do this by continuing here, I guess.
        if i in ignore_match_indices:
            continue

        new_matches.append(m)

    return new_matches, counts


def make_collection_json(
        template_file,
        output_dir,
        thresh,  # FIXME thresh not used in this version
        compress,
        ignore_match_indices=None):
    """
    Create a JSON collection file from a template file.

    Parameters
    ----------
    template_file : str
        Path to the template file.
    output_dir : str
        Directory where the output will be stored.
    thresh : float
        Threshold value.
    compress : bool
        Whether to compress the output.
    ignore_match_indices : Optional[List[int]], optional
        Indices of matches to ignore, by default None.

    Returns
    -------
    Tuple[str, List[Dict[str, int]]]
        A tuple containing the path to the collection file and a list of counts.
    """
    with open(template_file, 'r') as f:
        template_match_md = json.load(f)

    input_matches = template_match_md["collection"]

    m, counts = filter_match_collection(
        input_matches, thresh, 
        ignore_match_indices=ignore_match_indices
    )

    collection_file = os.path.join(output_dir, "collection.json")
    collection_file = jsongz.dump(m, collection_file, compress=compress)

    return collection_file, counts


class LensCorrectionSolver(ArgSchemaParser):
    default_schema = LensCorrectionSchema

    def __init__(self, *args, **kwargs):
        super(LensCorrectionSolver, self).__init__(*args, **kwargs)
        self.jtform = None

    def run(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.logger.setLevel(self.args['log_level'])
        utils.logger.setLevel(self.args['log_level'])
        self.check_for_files()
        self.output_dir = self.args.get('output_dir', self.args['data_dir'])
        self.logger.info("destination directory:\n  %s" % self.output_dir)

        tspecin = tilespec_input_from_metafile(
                self.metafile,
                self.args['mask_file'],
                self.output_dir,
                self.args['log_level'],
                self.args['compress_output'])
        gentspecs = GenerateEMTileSpecsModule(input_data=tspecin, args=[])
        gentspecs.run()

        assert os.path.isfile(gentspecs.args['output_path'])
        self.logger.info(
                "raw tilespecs written:\n  %s" % gentspecs.args['output_path'])

        collection_path, self.filter_counts = make_collection_json(
                self.matchfile,
                self.output_dir,
                self.args['ransac_thresh'],
                self.args['compress_output'],
                self.args['ignore_match_indices'])

        self.n_from_gpu = np.array(
                [i['n_from_gpu'] for i in self.filter_counts]).sum()
        self.n_after_filter = np.array(
                [i['n_after_filter'] for i in self.filter_counts]).sum()
        self.logger.info(
                "filter counts: %0.2f %% kept" %
                (100 * float(self.n_after_filter) / self.n_from_gpu))

        assert os.path.isfile(collection_path)
        self.logger.info(
                "filtered collection written:\n  %s" % collection_path)

        solver_args = {
                'nvertex': self.args['nvertex'],
                'regularization': self.args['regularization'],
                'good_solve': self.args['good_solve'],
                'tilespecs': gentspecs.tilespecs,
                'match_file': collection_path,
                'output_dir': self.output_dir,
                'outfile': 'resolvedtiles.json.gz',
                'compress_output': self.args['compress_output'],
                'log_level': self.args['log_level'],
                'timestamp': self.args['timestamp']}

        self.solver = MeshAndSolveTransform(input_data=solver_args, args=[])
        self.solver.run()

        with open(self.solver.args['output_json'], 'r') as f:
            j = json.load(f)
        resolved_path = j['resolved_tiles']

        resolvedtiles = renderapi.resolvedtiles.ResolvedTiles(
                json=jsongz.load(resolved_path))

        self.jtform = resolvedtiles.transforms[0].to_dict()

        self.map1, self.map2, self.mask = utils.maps_from_tform(
                renderapi.transform.ThinPlateSplineTransform(
                    json=self.jtform),
                resolvedtiles.tilespecs[0].width,
                resolvedtiles.tilespecs[0].height,
                res=32)

        maskname = os.path.join(self.output_dir, 'mask.png')
        cv2.imwrite(maskname, self.mask)
        self.logger.info("wrote:\n  %s" % maskname)

        res = {}
        res['input'] = {}
        res['input']['template'] = os.path.abspath(self.matchfile)
        res['input']['metafile'] = os.path.abspath(self.metafile)
        res['output'] = {}
        res['output']['resolved_tiles'] = j.pop('resolved_tiles')
        res['output']['mask'] = os.path.abspath(maskname)
        res['output']['collection'] = os.path.abspath(collection_path)
        res['residual stats'] = j

        self.args['output_json'] = self.solver.args['output_json']

        with open(self.args['output_json'], 'w') as f:
            json.dump(res, f, indent=2)

    def check_for_files(self):
        self.metafile = one_file(self.args['data_dir'], '_metadata*')
        self.matchfile = one_file(self.args['data_dir'], '_template_matches*')
        self.logger.info("Using files: \n  %s\n  %s" % (
            self.metafile, self.matchfile))


if __name__ == '__main__':
    lcmod = LensCorrectionSolver(input_data=example)
    lcmod.run()



import json
import os

from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import numpy as np

from argschema import ArgSchemaParser
from bigfeta import jsongz
import renderapi

from em_stitch.plots.schemas import MontagePlotsSchema

example = {
        "collection_path": "/data/em-131fs3/lctest/T4.2019.04.29b/001738/0/collection.json.gz",
        "resolved_path": "/data/em-131fs3/lctest/T4.2019.04.29b/001738/0/resolvedtiles_AffineModel_0.json.gz"
        }


def tspec_transform(tspec, mpq, shared=None):
    """
    Transform points using the transformations associated with a tilespec.
      Assumes any reference transforms are included in initial points and skips
      calculation for those.

    Parameters
    ----------
    tspec : renderapi.tilespec.TileSpec
        Tilespec object containing transformations.
    mpq : List[Tuple[float, float]]
        List of (x, y) points to transform.
    shared : Optional[Any], optional
        Shared information, by default None.

    Returns
    -------
    np.ndarray
        Transformed points.
    """
    xy = np.array(mpq).transpose()
    for tf in tspec.tforms:
        if isinstance(tf, renderapi.transform.ReferenceTransform):
            # matches already in LC units
            continue
        else:
            xy = tf.tform(xy)
    return xy


def make_xyres(matches, resolved):
    """
    Create XY coordinates and residuals from matches and resolved tiles.

    Parameters
    ----------
    matches : List[Dict[str, Any]]
        List of matches.
    resolved : renderapi.resolvedtiles.ResolvedTiles
        Resolved tiles object.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]
        Tuple containing XY coordinates and residuals for valid and invalid matches.
    """
    tids = np.array([t.tileId for t in resolved.tilespecs])
    xy = []
    res = []
    mxy = []
    mres = []
    for m in matches:
        pind = np.argwhere(m['pId'] == tids).flatten()
        qind = np.argwhere(m['qId'] == tids).flatten()
        if (pind.size != 1) & (qind.size != 1):
            continue
        pxy = tspec_transform(
                resolved.tilespecs[pind[0]],
                m['matches']['p'],
                shared=resolved.transforms)
        qxy = tspec_transform(
                resolved.tilespecs[qind[0]],
                m['matches']['q'],
                shared=resolved.transforms)
        w = np.array(m['matches']['w']) != 0.0
        xy.append(0.5 * (pxy[w] + qxy[w]))
        res.append(pxy[w] - qxy[w])
        w = np.invert(w)
        mxy.append(0.5 * (pxy[w] + qxy[w]))
        mres.append(pxy[w] - qxy[w])
    return (np.concatenate(xy), np.concatenate(res),
            np.concatenate(mxy), np.concatenate(mres))


def one_plot(f, ax, xy, c, vmin=-100, vmax=100, title=None, colorbar=True,fontsize=10):
    col = c
    if vmin == vmax:
        col = 'k'
    s = ax.scatter(
            xy[:, 0],
            xy[:, 1],
            c=col,
            marker='s',
            s=2.5,
            vmin=vmin,
            vmax=vmax)
    ax.set_aspect('equal')
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="3%", pad=0.5)
    if colorbar:
        f.colorbar(s, cax=cax)
    else:
        cax.axis('off')
    ax.set_title(title, fontsize=fontsize)


class MontagePlots(ArgSchemaParser):
    default_schema = MontagePlotsSchema

    def run(self):
        matches = jsongz.load(self.args['collection_path'])
        resolved = renderapi.resolvedtiles.ResolvedTiles(
                json=jsongz.load(self.args['resolved_path']))

        xy, res, mxy, mres = make_xyres(matches, resolved)
        if self.args['save_json_path']:
            with open(self.args['save_json_path'], 'w') as f:
                json.dump(
                        {
                            'xy': xy.tolist(),
                            'res': res.tolist(),
                            'filtered_xy': mxy.tolist(),
                            'filtered_res': mres.tolist()
                            }, f, indent=2)

        pdf = None 
        if self.args['pdf_out']:
            pdf = PdfPages(self.args['pdf_out'])
        if self.args['make_plot']:
            f, a = plt.subplots(
                    2, 2, clear=True, num=1,
                    sharex=True, sharey=True,
                    figsize=(12, 12))
            vmnmx = 5
            title = self.args['collection_path']
            title += '\n' + self.args['resolved_path'] + '\n'
            one_plot(
                    f, a[0, 0], xy, res[:, 0],
                    vmin=-vmnmx, vmax=vmnmx,
                    title=title+'x res [px]', fontsize=8)
            one_plot(
                    f, a[0, 1], xy, res[:, 1],
                    vmin=-vmnmx, vmax=vmnmx, title='y res [px]')
            one_plot(
                    f, a[1, 0], xy, np.linalg.norm(res, axis=1),
                    vmin=0, vmax=vmnmx*np.sqrt(2), title='mag res [px]')
            one_plot(
                    f, a[1, 1], mxy, np.linalg.norm(mres, axis=1),
                    vmin=0, vmax=vmnmx*np.sqrt(2), title='filtered')
            a[0, 0].invert_yaxis()
        if self.args['show']:
            plt.show()
        if pdf:
            pdf.savefig(f)
            pdf.close()
            print('wrote %s' % os.path.abspath(self.args['pdf_out']))


if __name__ == "__main__":
    mmod = MontagePlots(input_data=example)
    mmod.run()



import glob
import json
import os

from argschema import ArgSchemaParser
from bigfeta import jsongz
import bigfeta.bigfeta as bfa
import renderapi

from em_stitch.montage import meta_to_collection
from em_stitch.montage.schemas import MontageSolverSchema
from em_stitch.utils.generate_EM_tilespecs_from_metafile import (
    GenerateEMTileSpecsModule)
from em_stitch.utils.utils import pointmatch_filter, get_z_from_metafile

dname = os.path.join(
            os.path.dirname(os.path.abspath(__file__)),
            "../", "../", "integration_tests", "test_files")

example = {
        "data_dir": "/data/em-131fs3/lctest/T4_6/001844/0",
        "output_dir": "/data/em-131fs3/lctest/T4_6/001844/0",
        "ref_transform": None,
        "ransacReprojThreshold": 10,
        "solver_template_dir": dname,
        "solver_templates": [
            "affine_template.json",
            "polynomial_template.json",
            ]
        }


class MontageSolverException(Exception):
    pass


def do_solve(template_path, args, index):
    """
    Perform alignment solving based on the provided template and arguments.

    Parameters
    ----------
    template_path : str
        Path to the template file.
    args : Dict[str, Any]
        Arguments dictionary containing input and output information.
    index : int
        Index.

    Returns
    -------
    Dict[str, Any]
        Results of the alignment solving process.
    """
    with open(template_path, 'r') as f:
        template = json.load(f)
    template['input_stack']['input_file'] = \
        args['input_stack']['input_file']
    template['pointmatch']['input_file'] = \
        args['pointmatch']['input_file']
    template['output_stack']['compress_output'] = \
        args['output_stack']['compress_output']
    template['first_section'] = template['last_section'] = \
        args['first_section']
    fname = os.path.join(
            os.path.dirname(args['input_stack']['input_file']),
            'resolvedtiles_%s_%d.json' % (template['transformation'], index))
    template['output_stack']['output_file'] = fname
    template['fullsize_transform'] = False
    aligner = bfa.BigFeta(input_data=template, args=[])
    aligner.run()
    # these numbers only meaningful for fullsize_transform = False
    # to get results already in memory, on-scope, let's keep it that way
    # otherwise, we'll need a separate calculation that loads tilespecs
    # and matches to calculate residuals, costing more time
    res = {
            'output': os.path.basename(
                aligner.args['output_stack']['output_file']),
            'collection': os.path.basename(
                aligner.args['pointmatch']['input_file']),
            'x': {
                'mean': aligner.results['err'][0][0],
                'stdev': aligner.results['err'][0][1]
                },
            'y': {
                'mean': aligner.results['err'][1][0],
                'stdev': aligner.results['err'][1][1]
                },
            'mag': {
                'mean': aligner.results['mag'][0],
                'stdev': aligner.results['mag'][1]
                }
            }
    return res


def do_solves(collection, input_stack, z, compress, solver_args):
    """
    Perform multiple alignment solving processes.

    Parameters
    ----------
    collection : str
        Collection file.
    input_stack : str
        Input stack file.
    z : int
        Z value.
    compress : bool
        Whether to compress the output.
    solver_args : List[Dict[str, Any]]
        List of solver arguments.

    Returns
    -------
    List[Dict[str, Any]]
        List of results from alignment solving processes.
    """
    args = {'input_stack': {}, 'output_stack': {}, 'pointmatch': {}}
    args['input_stack']['input_file'] = input_stack
    args['pointmatch']['input_file'] = collection
    args['output_stack']['compress_output'] = compress
    args['first_section'] = args['last_section'] = z

    results = []
    for index, template in enumerate(solver_args):
        results.append(do_solve(template, args, index))

    return results


def montage_filter_matches(matches, thresh, model='Similarity'):
    """
    Filter matches in a montage.

    Parameters
    ----------
    matches : List[Dict[str, Any]]
        List of matches.
    thresh : float
        Threshold value.
    model : str, optional
        Model type, by default 'Similarity'.

    """
    for match in matches:
        _, _, w, _ = pointmatch_filter(
                match,
                n_clusters=1,
                n_cluster_pts=6,
                ransacReprojThreshold=thresh,
                model=model)
        match['matches']['w'] = w.tolist()


def get_metafile_path(datadir):
    """
    Get the path of the metadata file in the specified directory.

    Parameters
    ----------
    datadir : str
        Directory where the metadata file is located.

    Returns
    -------
    str
        Path of the metadata file.
    """
    return glob.glob(os.path.join(datadir, '_metadata*.json'))[0]


def make_raw_tilespecs(metafile, outputdir, groupId, compress):
    """
    Generate raw tilespecs from a metadata file.

    Parameters
    ----------
    metafile : str
        Path to the metadata file.
    outputdir : str
        Directory where the output will be stored.
    groupId : str
        Group ID.
    compress : bool
        Whether to compress the output.

    Returns
    -------
    Tuple[str, int]
        Path of the generated raw tilespecs file and the corresponding z value.

    """
    z = get_z_from_metafile(metafile)
    tspecin = {
            "metafile": metafile,
            "z": z,
            "sectionId": groupId,
            "output_path": os.path.join(outputdir, 'raw_tilespecs.json'),
            "compress_output": compress
            }
    gmod = GenerateEMTileSpecsModule(input_data=tspecin, args=[])
    gmod.run()
    return gmod.args['output_path'], z


def get_transform(metafile, tfpath, refdict, read_from):
    """
    Get transformation based on specified parameters.

    Parameters
    ----------
    metafile : str
        Path to the metadata file.
    tfpath : str
        Path to the transformation file.
    refdict : Dict[str, Any]
        Reference dictionary for transformation.
    read_from : str
        Source to read transformation data from ('metafile', 'reffile', or 'dict').

    Returns
    -------
    renderapi.transform.Transform
        Transformation object.
    """
    if read_from == 'metafile':
        with open(metafile, 'r') as f:
            j = json.load(f)
        tfj = j[2]['sharedTransform']
    elif read_from == 'reffile':
        with open(tfpath, 'r') as f:
            tfj = json.load(f)
    elif read_from == 'dict':
        tfj = refdict
    return renderapi.transform.Transform(json=tfj)


def make_resolved(rawspecpath, tform, outputdir, compress):
    """
    Generate resolved tiles from raw tilespecs and a transformation.

    Parameters
    ----------
    rawspecpath : str
        Path to the raw tilespecs file.
    tform : renderapi.transform.Transform
        Transformation object.
    outputdir : str
        Directory where the output will be stored.
    compress : bool
        Whether to compress the output.

    Returns
    -------
    str
        Path of the generated resolved tiles file.
    """
    # read in the tilespecs
    rtj = jsongz.load(rawspecpath)
    tspecs = [renderapi.tilespec.TileSpec(json=t) for t in rtj]

    # do not need this anymore
    os.remove(rawspecpath)

    # add the reference transform
    ref = renderapi.transform.ReferenceTransform()
    ref.refId = tform.transformId
    for t in tspecs:
        t.tforms.insert(0, ref)

    # make a resolved tile object
    resolved = renderapi.resolvedtiles.ResolvedTiles(
            tilespecs=tspecs,
            transformList=[tform])

    # write it to file and return the path
    rpath = os.path.join(outputdir, 'resolvedtiles_input.json')

    return jsongz.dump(resolved.to_dict(), rpath, compress)


class MontageSolver(ArgSchemaParser):
    default_schema = MontageSolverSchema

    def run(self):
        if 'metafile' not in self.args:
            self.args['metafile'] = get_metafile_path(self.args['data_dir'])
        else:
            self.args['data_dir'] = os.path.dirname(self.args['metafile'])

        if not self.args['output_dir']:
            self.args['output_dir'] = self.args['data_dir']

        # read the matches from the metafile
        matches = meta_to_collection.main([self.args['data_dir']])

        montage_filter_matches(
                matches,
                self.args['ransacReprojThreshold'])

        # write to file
        collection = os.path.join(self.args['output_dir'], "collection.json")
        collection = jsongz.dump(
                matches,
                collection,
                compress=self.args['compress_output'])

        # make raw tilespec json
        rawspecpath, z = make_raw_tilespecs(
                self.args['metafile'],
                self.args['output_dir'],
                matches[0]['pGroupId'],
                self.args['compress_output'])

        # get the ref transform
        tform = get_transform(
                self.args['metafile'],
                self.args['ref_transform'],
                self.args['ref_transform_dict'],
                self.args['read_transform_from'])

        # make a resolved tile object
        input_stack_path = make_resolved(
                rawspecpath,
                tform,
                self.args['output_dir'],
                self.args['compress_output'])

        templates = [os.path.join(self.args['solver_template_dir'], t)
                     for t in self.args['solver_templates']]
        self.results = do_solves(
                collection,
                input_stack_path,
                z,
                self.args['compress_output'],
                templates)

        self.args['output_json'] = os.path.join(
                self.args['output_dir'], 'montage_results.json')
        with open(self.args['output_json'], 'w') as f:
            json.dump(self.results, f, indent=2)


if __name__ == "__main__":
    mm = MontageSolver(input_data=example)
    mm.run()



import logging
import os

import cv2
import numpy as np
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt

from argschema import ArgSchemaParser
from bigfeta import jsongz
import renderapi

from em_stitch.plots.schemas import ViewMatchesSchema
from em_stitch.lens_correction.utils import maps_from_tform

logger = logging.getLogger()


def find_ind(matches, ids):
    # helper to find the index of matches
    # ind = find_ind(matches, ['_71_64', '_72_64'])
    for i in range(len(matches)):
        m = matches[i]
        pq = (ids[0] in m['pId']) & (ids[1] in m['qId'])
        qp = (ids[0] in m['qId']) & (ids[1] in m['pId'])
        if pq | qp:
            ind = i
    return ind


def get_ims_and_coords(m, ddir):
    pname = os.path.join(ddir, m['pId'] + '.tif')
    qname = os.path.join(ddir, m['qId'] + '.tif')
    pim = plt.imread(pname, 0)
    qim = plt.imread(qname, 0)
    p = np.array(m['matches']['p']).transpose()
    q = np.array(m['matches']['q']).transpose()
    w = np.array(m['matches']['w'])
    return pim, qim, p, q, w, pname, qname


def plot_ims_and_coords(
        pim, qim, p, q, w, pname=None, qname=None, fignum=1, tform=None):
    if tform:
        map1, map2, mask = maps_from_tform(
                tform, pim.shape[1], pim.shape[0])
        pim = cv2.remap(pim, map1, map2, cv2.INTER_NEAREST)
        qim = cv2.remap(qim, map1, map2, cv2.INTER_NEAREST)

    f, a = plt.subplots(1, 2, clear=True, num=fignum, figsize=(11, 8))

    a[0].imshow(pim, cmap='gray')
    nind = np.isclose(w, 0.0)
    ind = np.invert(nind)
    a[0].scatter(p[ind, 0], p[ind, 1], marker='x', color='r')
    a[0].scatter(p[nind, 0], p[nind, 1], marker='x', color='k')
    a[0].set_title(os.path.basename(pname), fontsize=8)

    a[1].imshow(qim, cmap='gray')
    a[1].scatter(q[ind, 0], q[ind, 1], marker='x', color='r')
    a[1].scatter(q[nind, 0], q[nind, 1], marker='x', color='k')
    a[1].set_title(os.path.basename(qname), fontsize=8)
    return f, a


example = {
        "data_dir": "/data/em-131fs3/lctest/T3_OL5pct/005270/0"
        }


class ViewMatches(ArgSchemaParser):
    default_schema = ViewMatchesSchema

    def run(self):
        cpath = self.args.get(
                'collection_path',
                os.path.join(
                    self.args['data_dir'],
                    self.args['collection_basename']))
        if (
                (not os.path.isfile(cpath)) &
                (os.path.splitext(cpath)[-1] == '.json')):
            cpath += '.gz'

        self.matches = jsongz.load(cpath)
        self.get_transform()

        if self.args['view_all']:
            inds = np.arange(len(self.matches))
        else:
            inds = [self.args['match_index']]

        with PdfPages(self.args['pdf_out']) as pdf:
            for ind in inds:
                pim, qim, p, q, w, pname, qname = get_ims_and_coords(
                        self.matches[ind],
                        self.args['data_dir'])
                f, a = plot_ims_and_coords(
                        pim, qim, p, q, w, pname=pname,
                        qname=qname, tform=self.tform, fignum=2)
                pdf.savefig(f)
                if self.args['show']:
                    plt.show()

        print('wrote %s' % os.path.abspath(self.args['pdf_out']))

    def get_transform(self):
        self.tform = None
        if 'transform_file' in self.args:
            self.tform = renderapi.transform.ThinPlateSplineTransform(
                    json=jsongz.load(self.args['transform_file']))
        else:
            for fbase in self.args['resolved_tiles']:
                fpath = os.path.join(
                        self.args['data_dir'],
                        fbase)
                if os.path.isfile(fpath):
                    resolved = renderapi.resolvedtiles.ResolvedTiles(
                            json=jsongz.load(fpath))
                    self.tform = resolved.transforms[0]
                    break


if __name__ == '__main__':
    vmod = ViewMatches(input_data=example)
    vmod.run()



import logging
import os

from argschema import ArgSchemaParser

from em_stitch.viz.schemas import SetUpdateUploadSchema
from em_stitch.viz.set_permissions import SetPermissions
from em_stitch.viz.update_urls import UpdateUrls
from em_stitch.viz.upload_to_render import UploadToRender

logger = logging.getLogger(__name__)


example = {
        # these 2 are how the local client (windows or posix) sees these files
        "fdir": "lctest/T3_OL5pct/005270/0",
        "collection_file": "collection.json.gz",
        "resolved_file": "resolvedtiles_AffineModel_0.json.gz",
        "server_mount": "/data/em-131fs3",  # leave as posix
        # "client_mount_or_map": "Q:/", # windows example
        "client_mount_or_map": "/data/em-131fs3",  # posix example
        "dir_setting": '777',
        'file_exts': ['.json', 'json.gz'],
        'file_setting': '777',
        "backup_copy": True,
        "render": {
                "host": "em-131fs",
                "port": 8987,
                "owner": "the_temcas",
                "project": "montage_test",
                "client_scripts": "/allen/aibs/pipeline/image_processing/volume_assembly/render-jars/production/scripts",
                "validate_client": False
              },
        "stack": "T3_example",
        "collection": "T3_example",
        "log_level": "INFO"
        }


def set_args(args):
    new_args = {k: args[k] for k in
                ['fdir', 'client_mount_or_map', 'log_level',
                 'dir_setting', 'file_exts', 'file_setting']}
    return dict(new_args)


def update_args(args):
    new_args = {k: args[k] for k in
                ['fdir', 'client_mount_or_map', 'log_level',
                 'backup_copy', 'resolved_file', 'image_directory']}
    return dict(new_args)


def upload_args(args):
    new_args = {k: args[k] for k in
                ['fdir', 'client_mount_or_map',
                 'render', 'stack', 'collection',
                 'collection_file', 'resolved_file', 'log_level']}
    return dict(new_args)


class SetUpdateUpload(ArgSchemaParser):
    default_schema = SetUpdateUploadSchema

    def run(self):
        logger.setLevel(self.args['log_level'])

        if os.name == 'posix':
            pset = SetPermissions(input_data=set_args(self.args), args=[])
            pset.run()

        fpaths = UpdateUrls(input_data=update_args(self.args), args=[])
        fpaths.run()

        up = UploadToRender(input_data=upload_args(self.args), args=[])
        up.run()


if __name__ == "__main__":
    s = SetUpdateUpload(input_data=example)
    s.run()
